from experimaestro.core.arguments import TypeAnnotation, ArgumentOptions


class Checker(TypeAnnotation):
    def check(self, value):
        """Check the value"""
        raise NotImplementedError()

    def annotate(self, options: ArgumentOptions):
        assert options.kwargs.get("checker", None) is None
        options.kwargs["checker"] = self


class Choices(Checker):
    def __init__(self, choices: list):
        self.choices = choices

    def check(self, value):
        for choice in self.choices:
            if value == choice:
                return True
        return False

    def __str__(self):
        return "value in " + ", ".join(str(choice) for choice in self.choices)



"""Dependency between tasks and tokens"""

import threading
from typing import Optional, Set, TYPE_CHECKING
import asyncio
from enum import Enum
from ..utils import logger
from ..locking import Lock

if TYPE_CHECKING:
    from . import Job


class Dependents:
    """Encapsulate the access to the dependents"""

    def __init__(self):
        self.lock = threading.Lock()
        self._dependents: Set[Dependency] = set()  # as source

    def add(self, dependency):
        with self.lock:
            self._dependents.add(dependency)

    def __enter__(self):
        # Returns the set after locking
        self.lock.__enter__()
        return self._dependents

    def __exit__(self, *args):
        self.lock.__exit__(*args)


class Resource:
    def __init__(self):
        self.dependents = Dependents()


class DependencyStatus(Enum):
    WAIT = 0
    """Waiting for dependency to be available"""

    OK = 1
    """Dependency can be locked"""

    FAIL = 2
    """Dependency won't be availabe in the foreseeable future"""


class Dependency:
    # Dependency status
    loop: asyncio.AbstractEventLoop

    def __init__(self, origin):
        # Origin and target are two resources
        self.origin = origin
        self.target: Optional["Job"] = None
        self.currentstatus = DependencyStatus.WAIT

    def status(self) -> DependencyStatus:
        raise NotImplementedError()

    def lock(self) -> Lock:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return "Dep[{origin}->{target}]/{currentstatus}".format(**self.__dict__)

    def check(self):
        assert self.target is not None
        status = self.status()
        logger.debug("Dependency check: %s", self)
        if status != self.currentstatus:
            logger.debug(
                "Dependency %s is %s (was: %s)", self, status, self.currentstatus
            )
            self.target.dependencychanged(self, self.currentstatus, status)
            self.currentstatus = status



"""mkdocs plugin for documentation generation

See https://www.mkdocs.org/user-guide/plugins/ for plugin API documentation
"""

from collections import defaultdict
import functools
import re
from experimaestro.mkdocs.annotations import shoulddocument
import requests
from urllib.parse import urljoin
from experimaestro.core.types import ObjectType, Type
import mkdocs
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Set, Tuple, Type as TypingType
import importlib
import logging
import itertools
import inspect
import mkdocs.config.config_options as config_options
from mkdocs.structure.pages import Page as MkdocPage
from experimaestro.core.objects import Config
import json
from docstring_parser.parser import parse as docstringparse

MODULEPATH = Path(__file__).parent


def md_protect(s):
    return re.sub(r"""([*`_{}[\]])""", r"""\\\1""", s)


def relativepath(source: str, target: str):
    """Computes a relative path from source to target"""
    if source == target:
        return ""

    if source[-1] == "/":
        source = source[:-1]

    if target[-1] == "/":
        target = target[:-1]

    source_seq = source.split("/")
    target_seq = target.split("/")
    maxlen = min(len(source_seq), len(target_seq))

    i = 0
    while i < maxlen and target_seq[i] == source_seq[i]:
        i += 1

    path = [".."] * (len(source_seq) - i) + target_seq[i:]

    return "/".join(path)


class ObjectLatticeNode:
    objecttype: ObjectType
    parents: Set["ObjectLatticeNode"]
    children: Set["ObjectLatticeNode"]

    def __init__(self, objecttype):
        self.objecttype = objecttype
        self.children = set()
        self.parents = set()

    def __hash__(self):
        return self.objecttype.__hash__()

    def __eq__(self, other):
        return other.objecttype is self.objecttype

    def __repr__(self):
        if self.objecttype is None:
            return "ROOT"
        return f"node({self.objecttype.identifier})"

    def isAncestor(self, other):
        return issubclass(self.objecttype.configtype, other.objecttype.configtype)

    def _addChild(self, child: "ObjectLatticeNode"):
        child.parents.add(self)
        self.children.add(child)

    def _removeChild(self, child: "ObjectLatticeNode"):
        child.parents.remove(self)
        self.children.remove(child)

    def add(self, node: "ObjectLatticeNode"):
        if self.objecttype == node.objecttype:
            assert id(self) == id(node)
            return

        added = False
        replace = set()
        for child in self.children:
            if node.isAncestor(child):
                # Add in child
                child.add(node)
                added = True
            elif child.isAncestor(node):
                # Replace child
                replace.add(child)

        # Replace children

        for child in replace:
            self._removeChild(child)
            node._addChild(child)
            self._addChild(node)

            added = True

        # No suitable parent found: add to ourself

        if not added:
            self._addChild(node)

        return node

    def find(self, objecttype):
        if objecttype == self.objecttype:
            return self

        for child in self.children:
            node = child.find(objecttype)
            if node is not None:
                return node

    def topologicalOrder(
        self, current: List["ObjectLatticeNode"], cover: Set["ObjectLatticeNode"]
    ):
        for child in self.children:
            current.append(child)

        for child in self.children:
            child.topologicalOrder(current, cover)

    def iter_all(self):
        return self._iter_all(set())

    def _iter_all(self, processed: Set):
        if id(self) not in processed:
            yield self
            processed.add(id(self))
            for node in self.children:
                yield from node._iter_all(processed)


class ObjectLattice:
    """Lattice of objects"""

    def __init__(self):
        self.node = ObjectLatticeNode(None)

    def add(self, objecttype: ObjectType):
        node = self.node.find(objecttype)
        if node is None:
            node = ObjectLatticeNode(objecttype)
            self.node.add(node)
        return node

    def topologicalOrder(self) -> List[ObjectLatticeNode]:
        current = []
        self.node.topologicalOrder(current, set())
        return current

    def iter_all(self) -> Iterator[ObjectLatticeNode]:
        return self.node.iter_all()


class Configurations:
    def __init__(self):
        self.tasks = set()
        self.configs = set()

    def add(self, node: ObjectLatticeNode):
        s = self.tasks if node.objecttype.task is not None else self.configs
        s.add(node)

    def remove(self, node):
        s = self.tasks if node.objecttype.task is not None else self.configs
        s.remove(node)

    def __iter__(self) -> Iterator[ObjectLatticeNode]:
        return itertools.chain(self.configs, self.tasks)


class Documentation(mkdocs.plugins.BasePlugin):
    RE_SHOWCLASS = re.compile(r"::xpm::([^ \n]+)")

    config_scheme = (
        ("name", config_options.Type(str, default="Tasks and configurations")),
        ("modules", config_options.Type(list)),
        ("external", config_options.Type(list)),
        ("init", config_options.Type(list)),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # path to sets of XPM types
        self.configurations: Dict[str, Configurations] = defaultdict(
            lambda: Configurations()
        )

        # Maps XPM types to markdown paths
        self.type2path = {}

        self.lattice = ObjectLattice()

    def on_config(self, config, **kwargs):
        # Import modules in init
        for module_name in self.config.get("init") or []:
            importlib.import_module(module_name)

        # Include documentation pages in config
        self.parsed = {}
        self.external = {}
        self.baseurl = config["site_url"]

        for item in self.config.get("external") or []:
            module_name, url = next(iter(item.items()))
            logging.info("Loading external mappings from %s", url)
            baseurl = str(urljoin(url, "."))
            mappings = requests.get(url).json()
            for module, path in mappings.items():
                self.external[module] = f"{baseurl}{path}"

        # Retrieve all configurations and tasks
        processed = set()
        for name_packagename in self.config["modules"]:
            md_path, module_names = next(iter(name_packagename.items()))
            path_cfgs = self.configurations[md_path]

            if md_path.endswith(".md"):
                md_path = md_path[:-3]

            for module_name in module_names:
                package = importlib.import_module(module_name)
                basepath = Path(package.__path__[0])

                for path in basepath.rglob("*.py"):
                    parts = list(path.relative_to(basepath).parts)
                    if parts[-1] == "__init__.py":
                        parts = parts[:-1]
                    elif parts[-1].endswith(".py"):
                        parts[-1] = parts[-1][:-3]

                    fullname = (
                        f"""{module_name}.{".".join(parts)}""" if parts else module_name
                    )

                    # Avoid to re-parse
                    if fullname in self.parsed:
                        continue
                    self.parsed[fullname] = f"{md_path}.html"

                    try:
                        module = importlib.import_module(fullname)
                        for _, member in inspect.getmembers(
                            module,
                            lambda t: inspect.isclass(t) and issubclass(t, Config),
                        ):
                            # Only include members of the module
                            if member.__module__ != fullname:
                                continue

                            xpmtype = member.__getxpmtype__()
                            if xpmtype in processed:
                                continue

                            processed.add(xpmtype)

                            node = self.lattice.add(xpmtype)
                            path_cfgs.add(node)

                            # Register on which page the type is defined
                            self.type2path[
                                f"{member.__module__}.{member.__qualname__}"
                            ] = f"{md_path}.html"

                            member.__xpmtype__.__initialize__()

                    except Exception as e:
                        logging.exception(
                            "Error while reading definitions file %s: %s", path, e
                        )
        return config

    def on_post_build(self, config):
        # Custom format export (FIXME: replace by sphinx)
        mapping_path = Path(config["site_dir"]) / "experimaestro-mapping.json"
        logging.info("Writing mapping file %s", mapping_path)
        with mapping_path.open("wt") as fp:
            json.dump(self.parsed, fp)

        # Sphinx export
        import zlib

        """Fron Sphinx source code"""

        def escape(string: str) -> str:
            return re.sub("\\s+", " ", string)

        path = Path(config["site_dir"]) / "objects.inv"

        with path.open("wb") as f:
            # header
            f.write(
                (
                    "# Sphinx inventory version 2\n"
                    "# Project: %s\n"
                    "# Version: %s\n"
                    "# The remainder of this file is compressed using zlib.\n"
                    % (escape("site_name"), escape("0.0.0"))
                ).encode()
            )

            # body
            compressor = zlib.compressobj(9)

            for node in self.lattice.iter_all():
                if node.objecttype is not None:
                    member = node.objecttype.basetype
                    qname = f"{member.__module__}.{member.__qualname__}"
                    path = self.type2path[qname]

                    f.write(
                        compressor.compress(
                            f"""{qname} py:xpmconfig 1 {path}#$ -\n""".encode("utf-8")
                        )
                    )

            f.write(compressor.flush())

    def showlink(self, location, m: re.Match):
        """Show a link to the documentation"""
        return self.getlink(location, m.group(1).strip())

    def showclass(self, location, m: re.Match, page: MkdocPage, cfgs: Configurations):
        """Show a class and its descendants"""

        # Search for the class

        modulename, classname = m.group(1).rsplit(":", 2)
        t = getattr(importlib.import_module(modulename), classname, None)
        if t is None:
            return f"<div class='error'>Cannot find {modulename}:{classname}</div>"
        if not issubclass(t, Config):
            return f"<div class='error'>{modulename}:{classname} is not an experimaestro Config(uration)</div>"

        xpmtype = t.__getxpmtype__()

        # Now, sort according to descendant/ascendant relationship or name
        nodes = set()
        for _node in cfgs:
            if issubclass(_node.objecttype.configtype, xpmtype.configtype):
                nodes.add(_node)

        # Removes so they are not generated twice
        for node in nodes:
            try:
                cfgs.remove(node)
            except Exception:
                logging.error("Cannot remove %s", node)

        # objecttypes.sort(key=functools.cmp_to_key(lambda a, b: if issubclass(a.objecttype, b.objecttype)))
        lines = []
        self.build_doc(page, lines, nodes)
        return "".join(lines)

    def _getlink(self, url, qualname: str) -> Tuple[str, Optional[str]]:
        """Get a link given a qualified name"""

        # Use an internal reference
        md_path = self.type2path.get(qualname, None)
        if md_path:
            return qualname, f"{relativepath(url, md_path)}#{qualname}"

        # Try to get an external reference
        module_name = qualname[: qualname.rfind(".")]
        baseurl = self.external.get(module_name, None)

        return qualname, f"{baseurl}#{qualname}" if baseurl else None

    def getlink(self, url, qualname: str):
        """Get a link given a qualified name"""
        qualname, href = self._getlink(url, qualname)
        if href:
            return f"[{qualname}]({href})"
        return qualname

    def getConfigLink(self, pageURL: str, config: TypingType):
        return self._getlink(pageURL, f"{config.__module__}.{config.__qualname__}")

    def build_doc(
        self, page: MkdocPage, lines: List[str], nodes: List[ObjectLatticeNode]
    ):
        """Build the documentation for a list of configurations"""

        # Sort
        # lattice = ObjectLattice()
        # for node in nodes:
        #     lattice.add(node.objecttype)

        # sortednodes = lattice.topologicalOrder()

        for node in nodes:
            xpminfo = node.objecttype
            fullqname = (
                f"{xpminfo.objecttype.__module__}.{xpminfo.objecttype.__qualname__}"
            )
            lines.extend(
                (
                    f"""### {xpminfo.title} <span id="{fullqname}"> </span>\n\n""",
                    f"""`from {xpminfo.objecttype.__module__} import {xpminfo.objecttype.__name__}`\n\n""",
                )
            )

            if xpminfo.description:
                lines.extend((xpminfo.description, "\n\n"))

            # Add parents
            parents = list(xpminfo.parents())
            if parents:
                lines.append("*Parents*: ")
                lines.append(
                    ", ".join(
                        self.getlink(page.url, parent.fullyqualifiedname())
                        for parent in parents
                    )
                )
                lines.append("\n\n")

            if node.children:
                lines.append("*Children*: ")
                lines.append(
                    ", ".join(
                        self.getlink(page.url, child.objecttype.fullyqualifiedname())
                        for child in node.children
                    )
                )
                lines.append("\n\n")

            for name, argument in xpminfo.arguments.items():

                if isinstance(argument.type, ObjectType):
                    basetype = argument.type.basetype
                    typestr = self.getlink(
                        page.url, f"{basetype.__module__}.{basetype.__qualname__}"
                    )
                else:
                    typestr = argument.type.name()

                lines.append("- ")
                if argument.generator:
                    lines.append(" [*generated*] ")
                elif argument.constant:
                    lines.append(" [*constant*] ")
                lines.append(f"**{name}** ({typestr})")
                if argument.help:
                    lines.append(f"\n  {argument.help}")
                lines.append("\n\n")

            methods = [
                member
                for key, member in inspect.getmembers(
                    xpminfo.objecttype,
                    predicate=lambda member: inspect.isfunction(member)
                    and shoulddocument(member),
                )
            ]

            if methods:
                lines.append("**Methods**\n\n")
                for method in methods:
                    parseddoc = docstringparse(method.__doc__)
                    lines.append(
                        f"""- {md_protect(method.__name__)}() *{parseddoc.short_description}*\n"""
                    )

    def on_page_markdown(self, markdown, page: MkdocPage, **kwargs):
        """Generate markdown pages"""
        path = page.file.src_path

        cfgs = self.configurations.get(path, None)
        if cfgs is None:
            return markdown

        markdown = Documentation.RE_SHOWCLASS.sub(
            lambda c: self.showclass(page.url, c, page, cfgs), markdown
        )

        lines = [
            markdown,
            "<style>",
            (MODULEPATH / "style.css").read_text(),
            "</style>\n",
            "<div><hr></div>",
            "*Documentation generated by experimaestro*\n",
        ]

        if cfgs.configs:
            lines.extend(["## Configurations\n\n"])
            self.build_doc(page, lines, cfgs.configs)

        if cfgs.tasks:
            lines.extend(["## Tasks\n\n"])
            self.build_doc(page, lines, cfgs.tasks)

        return "".join(lines)



"""Command line jobs"""

import json
import io
from pathlib import Path
from typing import List, Optional, Union, Callable, Dict
import itertools
from experimaestro.core.context import SerializationContext
from experimaestro.scheduler.workspace import RunMode

from experimaestro.utils import logger
from .scheduler import Job, JobState
from .connectors import Process, Redirect, RedirectType, Connector
from .scheduler import Workspace
from .core.objects import Config


# 5 seconds wait for locking file
LOCKFILE_WAIT_DURATION = 5


class NamedPipeRedirections:
    """List of named pipes"""

    def __init__(self):
        self.outputRedirections: List[Path] = []
        self.errorRedirections: List[Path] = []

    def redirections(self):
        return itertools.chain(self.outputRedirections, self.errorRedirections)


EMPTY_REDIRECTIONS = NamedPipeRedirections()


class CommandPart:
    def __init__(self):
        self.inputRedirect = Redirect.inherit()
        self.outputRedirect = Redirect.inherit()
        self.errorRedirect = Redirect.inherit()

    def forEach(self, f: Callable[["CommandPart"], None]):
        f(self)

    def output(self, context: "CommandContext", out: io.TextIOBase):
        raise NotImplementedError("output for %s" % self.__class__)


class AbstractCommandComponent(CommandPart):
    pass


class CommandContext(SerializationContext):
    def __init__(
        self,
        workspace: Workspace,
        connector: Connector,
        path: Path,
        name: str,
        config: Config,
    ):
        super().__init__()
        self.workspace = workspace
        self.connector = connector
        self.path = path
        self.name = name
        self.config = config
        self.namedPipeRedirectionsMap: Dict["CommandPart", NamedPipeRedirections] = {}
        self.auxiliary: Dict[str, int] = {}

    def getAuxiliaryFile(self, name, suffix):
        ix = self.auxiliary.get(name, 0) + 1
        self.auxiliary[name] = ix
        if ix == 1:
            return self.path / ("%s%s" % (name, suffix))
        return self.path / ("%s-%d%s" % (name, ix, suffix))

    def relpath(self, path):
        return self.connector.resolve(path, self.path)

    def getNamedRedirections(
        self, key: "CommandPart", create: bool
    ) -> NamedPipeRedirections:
        x = self.namedPipeRedirectionsMap.get(key, None)
        if x:
            return x

        if not create:
            return EMPTY_REDIRECTIONS

        x = NamedPipeRedirections()
        self.namedPipeRedirectionsMap[key] = x
        return x

    def writeRedirection(self, out, redirect, stream):
        raise NotImplementedError()

    def printRedirections(
        self, stream: int, out, outputRedirect: Redirect, outputRedirects
    ):
        raise NotImplementedError()


class CommandPath(AbstractCommandComponent):
    def __init__(self, path: Union[Path, str]):
        super().__init__()
        self.path = Path(path)

    def output(self, context, out):
        out.write(context.relpath(self.path))

    def __repr__(self):
        return "Path({})".format(self.path)


class CommandString(AbstractCommandComponent):
    def __init__(self, string: str):
        super().__init__()
        self.string = string

    def output(self, context, out):
        out.write(self.string)

    def __repr__(self):
        return "String({})".format(self.string)


class CommandParameters(AbstractCommandComponent):
    def output(self, context: CommandContext, out: io.TextIOBase):
        path = context.getAuxiliaryFile("params", ".json")
        with path.open("wt") as fileout:
            context.config.__xpm__.outputjson(fileout, context)
        out.write(context.relpath(path))


class AbstractCommand(CommandPart):
    def reorder(self):
        raise NotImplementedError()

    def output(self, context: CommandContext, out: io.TextIOBase):
        list = self.reorder()
        detached = 0

        if len(list) > 1:
            out.write("(\n")

        for command in list:
            # Write files
            namedRedirections = context.getNamedRedirections(command, False)

            # Write named pipes
            def mkfifo(file: Path):
                out.write(" mkfifo {}" % context.relpath(file))

            for file in namedRedirections.redirections():
                mkfifo(file)

            if command.inputRedirect.type == RedirectType.FILE:
                out.write(
                    " cat {} | ".format(context.relpath(command.inputRedirect.path))
                )

            command.output(context, out)

            context.printRedirections(
                1, out, command.outputRedirect, namedRedirections.outputRedirections
            )
            context.printRedirections(
                2, out, command.errorRedirect, namedRedirections.errorRedirections
            )

            out.write('|| checkerror "${PIPESTATUS[@]}" || exit $?')

        # Monitors detached jobs
        for i in range(detached):
            out.write("wait $CHILD_{} || exit $?%\n".format(i))

        if len(list) > 1:
            out.write(")\n")


class Command(AbstractCommand):
    def __init__(self):
        super().__init__()
        self.components = []

    def add(self, component: AbstractCommandComponent):
        self.components.append(component)

    def __repr__(self):
        return "Command({})".format(",".join(str(c) for c in self.components))

    def output(self, context, out):
        first = True
        for c in self.components:
            if first:
                first = False
            else:
                out.write(" ")
            c.output(context, out)

    def forEach(self, f: Callable[["CommandPart"], None]):
        f(self)
        for c in self.components:
            c.forEach(f)


class CommandLine(AbstractCommand):
    """A command line is composed of one or more commands"""

    def __init__(self):
        super().__init__()
        self.commands = []

    def add(self, command: Command):
        self.commands.append(command)

    def forEach(self, f: Callable[["CommandPart"], None]):
        f(self)
        for command in self.commands:
            command.forEach(f)

    def reorder(self):
        return self.commands


class CommandLineJob(Job):
    def __init__(
        self,
        commandline: CommandLine,
        parameters,
        workspace: Optional[Workspace] = None,
        launcher=None,
        run_mode: RunMode = None,
    ):
        super().__init__(
            parameters, workspace=workspace, launcher=launcher, run_mode=run_mode
        )
        self.commandline = commandline

    async def aio_process(self) -> Optional[Process]:
        """Returns the process if there is one"""
        if self._process:
            return self._process

        if self.pidpath.is_file():
            # Get from pidpath file
            from experimaestro.connectors import Process

            pinfo = json.loads(self.pidpath.read_text())
            p = Process.fromDefinition(self.launcher.connector, pinfo)
            if p is None:
                return None

            if await p.aio_isrunning():
                return p

            return None

        return None

    @property
    def notificationURL(self):
        if self.launcher and self.launcher.notificationURL:
            return self.launcher.notificationURL
        return self.workspace.notificationURL

    def prepare(self, overwrite=False):
        """Prepare all files before starting a task

        :param overwrite: if True, overwrite files even if the task has been run
        """
        logger.debug("Preparing job %s...", self)

        assert self.launcher is not None, "No launcher defined for this job"

        scriptbuilder = self.launcher.scriptbuilder()
        self.path.mkdir(parents=True, exist_ok=True)
        donepath = self.donepath

        # Check again if done (now that we have locked)
        if not overwrite and donepath.is_file():
            logger.info("Job %s is already done", self)
            return JobState.DONE

        # Now we can write the script
        scriptbuilder.lockfiles.append(self.lockpath)
        scriptbuilder.command = self.commandline
        scriptbuilder.notificationURL = self.notificationURL
        return scriptbuilder.write(self)

    async def aio_run(self):
        if self._process:
            return self._process

        scriptPath = self.prepare()

        logger.info("Starting job %s", self.jobpath)
        processbuilder = self.launcher.processbuilder()
        processbuilder.environ = self.environ
        processbuilder.command.append(self.launcher.connector.resolve(scriptPath))
        processbuilder.stderr = Redirect.file(self.stderr)
        processbuilder.stdout = Redirect.file(self.stdout)
        self._process = processbuilder.start()

        with self.pidpath.open("w") as fp:
            json.dump(self._process.tospec(), fp)

        self.state = JobState.RUNNING
        logger.info("Process started (%s)", self._process)
        return self._process


class CommandLineTask:
    def __init__(self, commandline: CommandLine):
        self.commandline = commandline

    def __call__(
        self, pyobject, *, launcher=None, workspace=None, run_mode=None
    ) -> Job:
        return CommandLineJob(
            self.commandline,
            pyobject,
            launcher=launcher,
            workspace=workspace,
            run_mode=run_mode,
        )



import tempfile
import shutil
import os
from pathlib import Path
import logging
import signal

from experimaestro import experiment, task
from experimaestro.scheduler.workspace import RunMode


class TimeInterval:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __lt__(self, value):
        return self.start > value.end

    def __str__(self):
        return "%.4f - %.4f" % (self.start, self.end)

    def __repr__(self):
        return str(self)


def get_times(task: task) -> TimeInterval:
    logging.info("Reading times from %s", task.stdout())
    return TimeInterval(
        *(float(t) for t in task.stdout().read_text().strip().split("\n"))
    )


def get_times_frompath(path) -> TimeInterval:
    s = path.read_text().strip().split("\n")
    logging.info("Read times: %s", s)
    return TimeInterval(*(float(t) for t in s))


class TemporaryDirectory:
    def __init__(self, suffix=None, prefix=None, dir=None):
        self.suffix = suffix
        self.prefix = prefix
        self.dir = dir
        self.path = None

    def __enter__(self):
        self.path = Path(
            tempfile.mkdtemp(suffix=self.suffix, prefix=self.prefix, dir=self.dir)
        )
        return self.path

    def __exit__(self, exc_type, exc_value, traceback):
        if os.environ.get("XPM_KEEPWORKDIR", False) == "1":
            logging.warning("NOT Removing %s" % self.path)
        else:
            logging.warning(
                "Cleaning up working directory %s"
                " (use XPM_KEEPWORKDIR=1 to keep it)",
                self.path,
            )
            shutil.rmtree(self.path, ignore_errors=True)


class timeout:
    def __init__(self, seconds, error_message=None):
        if error_message is None:
            error_message = "test timed out after {}s.".format(seconds)
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        logging.error("Timeout - sending signal")
        import faulthandler

        faulthandler.dump_traceback()
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)


class TemporaryExperiment:
    def __init__(
        self,
        name,
        workdir=None,
        maxwait=20,
        port=None,
        run_mode: RunMode = RunMode.NORMAL,
    ):
        self.name = name
        self.workdir = workdir
        self.clean_workdir = workdir is None
        self.timeout = timeout(maxwait) if maxwait > 0 else 0
        self.port = port
        self.run_mode = run_mode

    def __enter__(self) -> experiment:
        if self.clean_workdir:
            self.workdir = TemporaryDirectory(prefix="xpm", suffix=self.name)
            workdir = self.workdir.__enter__()
        else:
            workdir = self.workdir

        self.experiment = experiment(
            workdir, self.name, port=self.port, run_mode=self.run_mode
        )
        self.experiment.__enter__()

        # Set some useful environment variables
        self.experiment.workspace.launcher.setenv(
            "PYTHONPATH", str(Path(__file__).parents[2])
        )
        if self.timeout:
            self.timeout.__enter__()

        logging.info("Created new temporary experiment (%s)", workdir)
        return self.experiment

    def __exit__(self, *args):
        self.experiment.__exit__(*args)
        if self.timeout:
            self.timeout.__exit__(*args)
        if self.clean_workdir:
            self.workdir.__exit__(*args)


def is_posix():
    try:
        import posix  # noqa: F401

        return True
    except ImportError:
        return False



if __name__ == "__main__":
    import sys
    import logging
    from pathlib import Path
    import time

    from experimaestro.scheduler import JobState
    from experimaestro.tests.utils import (
        TemporaryExperiment,
    )
    from experimaestro.tests.task_tokens import TokenTask

    root = logging.getLogger()
    root.setLevel(logging.INFO)
    logging.getLogger("xpm").setLevel(logging.DEBUG)

    workdir, x, lockingpath, readypath, timepath = sys.argv[1:]

    handler = logging.StreamHandler()
    bf = logging.Formatter(
        f"[XP{x}] "
        "[%(levelname)s] %(asctime)s %(name)s "
        "[%(process)d/%(threadName)s]: %(message)s",
        datefmt="%H:%M:%S.%f",
    )
    handler.setFormatter(bf)
    root.handlers.clear()
    root.addHandler(handler)

    with TemporaryExperiment("reschedule%s" % x, workdir=workdir) as xp:
        logging.info("Reschedule with token [%s]: starting task in %s", x, workdir)
        token = xp.workspace.connector.createtoken("test-token-reschedule", 1)
        task = (
            TokenTask(path=lockingpath, x=int(x))
            .add_dependencies(token.dependency(1))
            .submit()
        )
        logging.info("Waiting for task (token with %s) to be scheduled", lockingpath)
        while task.job.state == JobState.UNSCHEDULED:
            time.sleep(0.01)

        # Write so that the test now we are ready
        Path(readypath).write_text("hello")
        logging.info("Reschedule with token [%s]: ready", x)

        # Wait until the experiment
        task.__xpm__.task.job.wait()
        logging.info("Reschedule with token [%s]: finished", x)
        Path(timepath).write_text(Path(task.stdout()).read_text())



import logging
from attr import Factory
from attrs import define
from copy import copy, deepcopy
from dataclasses import dataclass
from typing import List, Optional, Union
from humanfriendly import parse_size, format_size, parse_timespan

# --- Host specification part


@dataclass
class CudaSpecification:
    memory: int
    """Memory (in bytes)"""

    model: str = ""
    """CUDA card model name"""

    min_memory: int = 0
    """Minimum request memory (in bytes)"""

    def __lt__(self, other: "CudaSpecification"):
        return self.memory < other.memory

    def match(self, spec: "CudaSpecification"):
        """Returns True if the specification matches this host"""
        return (self.memory >= spec.memory) and (self.min_memory <= spec.memory)

    def __repr__(self):
        return (
            f"CUDA({self.model} "
            f"{format_size(self.memory)}/{format_size(self.min_memory)})"
        )


@dataclass
class CPUSpecification:
    memory: int = 0
    """Memory in bytes"""

    cores: int = 0
    """Number of cores"""

    mem_per_cpu: int = 0
    """Memory per CPU (0 if not defined)"""

    cpu_per_gpu: int = 0
    """Number of CPU per GPU (0 if not defined)"""

    def __lt__(self, other: "CPUSpecification"):
        return self.memory < other.memory and self.cores < other.cores

    def total_memory(self, gpus: int = 0):
        return max(
            self.memory,
            self.mem_per_cpu * self.cores,
            self.cpu_per_gpu * self.mem_per_cpu * gpus,
        )


@define(kw_only=True)
class HostSpecification:
    """Specifies how the host is set"""

    cuda: List[CudaSpecification] = Factory(list)
    """CUDA GPUs"""

    cpu: CPUSpecification = Factory(CPUSpecification)
    """CPU specification for this host"""

    priority: int = 0
    """Priority for this host (higher better)"""

    max_duration: int = 0
    """Max job duration (in seconds)"""

    min_gpu: int = 0
    """Minimum number of allocated GPUs"""

    def __post_init__(self):
        self.cuda = sorted(self.cuda)


# --- Query part


@dataclass
class MatchRequirement:
    score: float
    requirement: "HostSimpleRequirement"


class HostRequirement:
    """A requirement must be a disjunction of host requirements"""

    requirements: List["HostSimpleRequirement"]

    def __init__(self) -> None:
        self.requirements = []

    def __or__(self, other: "HostRequirement"):
        return RequirementUnion(self, other)

    def match(self, host: HostSpecification) -> Optional[MatchRequirement]:
        raise NotImplementedError()


class RequirementUnion(HostRequirement):
    """Ordered list of simple host requirements -- the first one is the priority"""

    requirements: List["HostSimpleRequirement"]

    def __init__(self, *requirements: "HostSimpleRequirement"):
        self.requirements = list(requirements)

    def match(self, host: HostSpecification) -> Optional[MatchRequirement]:
        """Returns the matched requirement (if any)"""

        argmax: Optional[MatchRequirement] = None

        for req in self.requirements:
            max_score = float("-inf") if argmax is None else argmax.score

            if match := req.match(host):
                if match.score > max_score:
                    argmax = MatchRequirement(match.score, req)

        return argmax


class HostSimpleRequirement(HostRequirement):
    """Simple host requirement"""

    cuda_gpus: List["CudaSpecification"]
    cpu: "CPUSpecification"

    duration: int
    """Requested duration (in seconds)"""

    def __repr__(self):
        return f"Req(cpu={self.cpu}, cuda={self.cuda_gpus}, duration={self.duration})"

    def __init__(self, *reqs: "HostSimpleRequirement"):
        self.cuda_gpus = []
        self.cpu = CPUSpecification(0, 0)
        self.duration = 0
        for req in reqs:
            self._add(req)

    def __and__(self, other: "HostSimpleRequirement"):
        newself = copy(self)
        newself._add(other)
        return newself

    def _add(self, req: "HostSimpleRequirement"):
        self.cpu.memory = max(req.cpu.memory, self.cpu.memory)
        self.cpu.cores = max(req.cpu.cores, self.cpu.cores)
        self.duration = max(req.duration, self.duration)
        self.cuda_gpus.extend(req.cuda_gpus)
        self.cuda_gpus.sort()

    def match(self, host: HostSpecification) -> Optional[MatchRequirement]:
        if self.cuda_gpus:
            if len(host.cuda) < len(self.cuda_gpus):
                logging.debug(
                    "Not enough CUDA gpus (%d < %d)",
                    len(host.cuda),
                    len(self.cuda_gpus),
                )
                return None

            for host_gpu, req_gpu in zip(host.cuda, self.cuda_gpus):
                if not host_gpu.match(req_gpu):
                    return None

        if len(self.cuda_gpus) < host.min_gpu:
            logging.debug(
                "Not enough requested CUDA gpus (min=%d > %d)",
                host.min_gpu,
                len(self.cuda_gpus),
            )
            return None

        if host.cpu < self.cpu:
            return None

        if host.max_duration > 0 and self.duration > host.max_duration:
            return None

        return MatchRequirement(host.priority, self)

    def __mul__(self, count: int) -> "HostSimpleRequirement":
        if count == 1:
            return self

        _self = deepcopy(self)
        for _ in range(count - 1):
            _self.cuda_gpus.extend(self.cuda_gpus)
        _self.cuda_gpus.sort()

        return _self


def cpu(*, mem: Optional[str] = None, cores: int = 1):
    """CPU requirement"""
    r = HostSimpleRequirement()
    r.cpu = CPUSpecification(parse_size(mem) if mem else 0, cores)
    return r


def cuda_gpu(*, mem: Optional[str] = None):
    """CUDA GPU requirement"""
    _mem = parse_size(mem) if mem else 0
    r = HostSimpleRequirement()
    r.cuda_gpus.append(CudaSpecification(_mem))
    return r


def duration(timespec: Union[str, int]):
    """Request a given time duration


    :param timespec: A string like 5h (5 hours), 10m (10 minutes) or 42s (42
        seconds). parsable by
        [humanfriendly](https://humanfriendly.readthedocs.io/en/latest/api.html)
        or a number of seconds
    """
    r = HostSimpleRequirement()
    if isinstance(timespec, (int, float)):
        r.duration = int(timespec)
    else:
        r.duration = int(parse_timespan(timespec))

    return r



from experimaestro import argument, Task


@argument("value", type=int)
class IntegerTask(Task):
    def execute(self):
        if not isinstance(self.value, int):
            raise AssertionError("Not an integer")


@argument("value", type=float)
class FloatTask(Task):
    def execute(self):
        if not isinstance(self.value, float):
            raise AssertionError("Not a float but %s" % type(self.value))



from pathlib import Path
from typing import Optional
import os
import logging


class Env:
    _instance = None

    # The working directory path
    wspath: Optional[Path] = None

    # The current task path
    taskpath: Optional[Path] = None

    # Set to True when multi-processing when
    # in slave mode:
    # - no progress report
    slave: bool = False

    @staticmethod
    def instance():
        if Env._instance is None:
            Env._instance = Env()
        return Env._instance



"""Utilities exposed to users of the experimaestro API"""

from pathlib import Path
from experimaestro.core.objects import ConfigWalkContext


class DirectoryContext(ConfigWalkContext):
    """Special generation context used for debugging and testing"""

    def __init__(self, path: Path):
        super().__init__()
        self._path = Path(path)

    @property
    def path(self):
        return self._path


class EmptyContext(ConfigWalkContext):
    """Special generation context used for debugging and testing"""

    @property
    def path(self):
        raise AssertionError("Empty experimaestro context does not define a path")



"""Management of the arguments (params, options, etc) associated with the XPM objects"""

from typing import Optional, TypeVar, TYPE_CHECKING
from experimaestro.typingutils import get_optional
from pathlib import Path
import sys

if TYPE_CHECKING:
    from typing_extensions import Annotated
    import experimaestro.core.types
else:
    if sys.version_info.major == 3 and sys.version_info.minor < 9:
        from typing_extensions import Annotated
    else:
        from typing import Annotated


class Argument:
    """Represents an argument of a configuration or task"""

    objecttype: Optional["experimaestro.core.types.ObjectType"]
    """The object for which this argument was declared"""

    def __init__(
        self,
        name,
        type: "experimaestro.core.types.Type",
        required=None,
        help=None,
        generator=None,
        ignored=None,
        default=None,
        checker=None,
        constant=False,
        is_data=False,
    ):
        """Creates a new argument

        Args:
            name (str): The name of the argument

            type (experimaestro.core.types.Type): The type of the argument

            required (bool, optional): True if required (if None, determines
            automatically). Defaults to None.

            help (str, optional): Help string. Defaults to None.

            generator (Generator, optional): The value generator (e.g. for
            paths). Defaults to None.

            ignored (bool, optional): True if ignored (if None, computed
            automatically). Defaults to None.

            default (any, optional): . Defaults to None.

            checker (any, optional): Value checker. Defaults to None.

            constant (bool, optional): If true, the value is constant. Defaults
            to False.

            is_data (bool, optional): Flag for paths that are data path (to be
            serialized). Defaults to False.
        """
        required = (default is None) if required is None else required
        if default is not None and required is not None and required:
            raise Exception(
                "argument '%s' is required but default value is given" % name
            )

        self.name = name
        self._help = help
        self.checker = checker
        self.type = type
        self.constant = constant
        self.ignored = self.type.ignore if ignored is None else ignored
        self.required = required
        self.default = default
        self.generator = generator
        self.objecttype = None
        self.is_data = is_data

        assert (
            not self.constant or self.default is not None
        ), "Cannot be constant without default"

    def __repr__(self):
        return "Param[{name}:{type}]".format(**self.__dict__)

    def validate(self, value):
        value = self.type.validate(value)
        if self.checker:
            if not self.checker.check(value):
                raise ValueError("Value %s is not valid", value)
        return value

    def isoutput(self):
        if self.generator:
            return self.generator.isoutput()
        return False

    @property
    def help(self):
        if self._help is None and self.objecttype is not None:
            self.objecttype.__parsedoc__()
        return self._help

    @help.setter
    def help(self, help: str):
        self._help = help


class ArgumentOptions:
    """Helper class when using type hints"""

    def __init__(self):
        self.kwargs = {}
        self.constant = False

    def create(self, name, originaltype, typehint):
        from experimaestro.core.types import Type

        optionaltype = get_optional(typehint)
        type = Type.fromType(optionaltype or typehint)

        if "default" not in self.kwargs or self.kwargs["default"] is None:
            defaultvalue = getattr(originaltype, name, None)
            self.kwargs["default"] = defaultvalue

        self.kwargs["required"] = (optionaltype is None) and (
            self.kwargs["default"] is None
        )

        return Argument(name, type, **self.kwargs)


class TypeAnnotation:
    def __call__(self, options: Optional[ArgumentOptions]):
        if options is None:
            options = ArgumentOptions()
        self.annotate(options)
        return options

    def annotate(self, options: ArgumentOptions):
        pass


class _Param(TypeAnnotation):
    """Base annotation for types"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def annotate(self, options: ArgumentOptions):
        options.kwargs.update(self.kwargs)
        return options


T = TypeVar("T")

paramHint = _Param()
Param = Annotated[T, paramHint]

optionHint = _Param(ignored=True)
Option = Annotated[T, optionHint]
Meta = Annotated[T, optionHint]

dataHint = _Param(ignored=True, is_data=True)
DataPath = Annotated[Path, dataHint]
"""Annotates a path that should be kept to restore an object to its state"""


class help(TypeAnnotation):
    def __init__(self, text: str):
        self.text = text

    def annotate(self, options: ArgumentOptions):
        options.kwargs["help"] = self.text


class default(TypeAnnotation):
    """Adds a default value (useful when we have problems with setattr and class
    properties)"""

    def __init__(self, value):
        self.value = value

    def annotate(self, options: ArgumentOptions):
        options.kwargs["default"] = self.value


class ConstantHint(TypeAnnotation):
    def annotate(self, options: ArgumentOptions):
        options.kwargs["constant"] = True


constantHint = ConstantHint()
Constant = Annotated[T, constantHint]



from typing import Any, Callable
import pytest
from experimaestro import Config, Param, Task, RunMode
from experimaestro.scheduler.base import JobDependency
from experimaestro.tests.utils import TemporaryExperiment


@pytest.fixture()
def xp():
    with TemporaryExperiment("deprecated", maxwait=0, run_mode=RunMode.DRY_RUN) as xp:
        yield xp


def check_dependencies(task: Task, *tasks: Task):
    deps = [
        id(dep.origin.config)
        for dep in task.__xpm__.job.dependencies
        if isinstance(dep, JobDependency)
    ]

    assert len(deps) == len(tasks)
    for task in tasks:
        assert (
            id(task) in deps
        ), f"Task {task.__xpmtype__} with ID {task.__identifier__()}"
        " is not not in the dependencies"


class TaskA(Task):
    pass


class TaskB(Task):
    a: Param[TaskA]


def test_dependencies_simple(xp):
    a = TaskA().submit()
    b = TaskB(a=a).submit()
    check_dependencies(b, a)


def test_dependencies_implicit(xp):
    a = TaskA().submit()
    b = TaskB(a=a)
    b.submit()
    check_dependencies(b, a)


class TaskC(Task):
    def task_outputs(self, dep: Callable[[Config], None]) -> Any:
        return dep(ConfigC(param_c=self))


class ConfigC(Config):
    param_c: Param[TaskC]


class TaskD(Task):
    param_c: Param[ConfigC]


def test_dependencies_task_output(xp):
    task_c = TaskC()
    c = task_c.submit()
    d = TaskD(param_c=c).submit()
    check_dependencies(d, task_c)


class Inner_TaskA(Task):
    def task_outputs(self, dep: Callable[[Config], None]) -> Any:
        return dep(Inner_OutputTaskA())


class Inner_OutputTaskA(Config):
    pass


class Inner_TaskB(Task):
    param_a: Param[Inner_OutputTaskA]


def test_dependencies_inner_task_output(xp):
    task_a = Inner_TaskA()
    a = task_a.submit()
    b = Inner_TaskB(param_a=a).submit()
    check_dependencies(b, task_a)


def test_dependencies_pre_task(xp):
    a = TaskA().submit()
    a2 = TaskA().add_pretasks(a).submit()
    check_dependencies(a2, a)



from dataclasses import dataclass
from pathlib import Path, _posix_flavour
import io
import os
import re
from experimaestro.launcherfinder import LauncherRegistry, YAMLDataClass
from fabric import Connection
from invoke import Promise
import invoke.exceptions
from urllib.parse import urlparse
from itertools import chain
from . import Connector
from . import (
    Process,
    ProcessBuilder,
    RedirectType,
    Redirect,
)
from experimaestro.locking import Lock
from experimaestro.tokens import Token

# Might be wise to switch to https://github.com/marian-code/ssh-utilities


class SshPath(Path):
    """SSH path

    Absolute:
    ssh://[user@]host[:port]//this/is/a/path

    Relative:
    ssh://[user@]host[:port]/relative/path
    """

    _flavour = _posix_flavour

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, url: str):
        parsed = urlparse(url)
        assert parsed.scheme == "ssh"
        self._host = parsed.hostname

        self._parts = re.split(r"/+", parsed.path)
        if parsed.path.startswith("//"):
            self._parts[0] = "/"
        else:
            self._parts = self._parts[1:]

    @property
    def hostpath(self):
        if self.is_absolute():
            return "/" + "/".join(self._parts[1:])
        return "/".join(self._parts)

    @property
    def host(self):
        return self._host

    def is_absolute(self):
        return self._parts and self._parts[0] == "/"

    @classmethod
    def _parse_args(cls, args):
        drv = ""
        if args[0].startswith("ssh:"):
            url = urlparse(args[0])
            assert not url.fragment and not url.query

            path = url.path

            if path.startswith("//"):
                # Absolute path
                args = tuple(chain(["/", path[2:]], args[1:]))
            else:
                args = tuple(chain([path[1:]], args[1:]))

            drv = url.hostname
        _, root, parts = super()._parse_args(args)
        return (drv, root, parts)

    def _make_child(self, args):
        drv, root, parts = self._parse_args(args)
        assert self._host == drv or drv == "", f"{self._host} and {drv}"
        drv, root, parts = self._flavour.join_parsed_parts(
            "", "", self._parts, "", root, parts
        )

        child = object.__new__(SshPath)
        child._parts = parts
        child._host = self._host
        return child

    def open(self, mode="r", buffering=-1, encoding=None, errors=None, newline=None):
        # FIXME: should probably be wiser
        fileobj = (
            SshConnector.get(self.host)
            .connection.sftp()
            .open(self.hostpath, mode, buffering)
        )
        if "b" in mode:
            return fileobj

        return io.TextIOWrapper(
            fileobj, encoding=encoding, newline=newline, errors=errors
        )

    def absolute(self):
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.

        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        # XXX untested yet!
        if self._closed:
            self._raise_closed()
        if self.is_absolute():
            return self
        # FIXME this must defer to the specific flavour (and, under Windows,
        # use nt._getfullpathname())
        obj = self._from_parts([os.getcwd()] + self._parts, init=False)
        obj._init(template=self)
        return obj

    def __repr__(self):
        return "SshPath(%s,%s)" % (self._host, self._flavour.join(self._parts[1:]))

    def __str__(self):
        return "ssh://%s/%s" % (self._host, self._flavour.join(self._parts[1:]))


@dataclass
class SshConfiguration(YAMLDataClass):
    hostname: str

    def create(self, registry: LauncherRegistry):
        return SshConnector.get(self.hostname)


def get_stream(redirect: Redirect, write: bool):
    if redirect.type == RedirectType.FILE:
        raise NotImplementedError()

    if redirect.type == RedirectType.PIPE:
        raise NotImplementedError()

    if redirect.type == RedirectType.INHERIT:
        return None

    raise NotImplementedError("For %s", redirect)


class SshProcess(Process):
    def __init__(self, promise: Promise):
        self.promise = promise

    def __repr__(self):
        return f"SshProcess({self._process.pid})"

    def tospec(self):
        return {"type": "ssh", "pid": self._process.pid}

    def kill(self):
        self._process.kill()

    @staticmethod
    def fromspec(connector, spec):
        return None

    def wait(self) -> int:
        try:
            self.promise.join()
        except invoke.exceptions.Failure:
            raise


class SshProcessBuilder(ProcessBuilder):
    def __init__(self, connector: "SshConnector"):
        super().__init__()
        self.connector = connector

    def start(self):
        """Start the process"""

        trans = str.maketrans({'"': r"\"", "$": r"\$"})
        command = f'''"{'", "'.join([c.translate(trans) for c in self.command])}"'''

        # stdin = get_stream(self.stdin, False)
        # stdout = get_stream(self.stdout, True)
        # stderr = get_stream(self.stderr, True)

        self.connector.connection.run(
            command, asynchronous=not self.detach, disown=self.detach
        )
        raise NotImplementedError()
        # if self.detach:
        #     self.connector.connection
        #     p = subprocess.Popen(
        #         self.command,
        #         stdin=stdin,
        #         stderr=stderr,
        #         stdout=stdout,
        #         env=self.environ,
        #         close_fds=True,
        #         cwd="/",
        #     )
        # else:
        #     p = subprocess.Popen(
        #         self.command,
        #         stdin=stdin,
        #         stderr=stderr,
        #         stdout=stdout,
        #         env=self.environ,
        #     )

        # process = LocalProcess(p)

        # if self.stdout and self.stdout.type == RedirectType.PIPE:
        #     self.stdout.function(p.stdout)
        # if self.stderr and self.stderr.type == RedirectType.PIPE:
        #     self.stderr.function(p.stderr)

        # return process


class SshConnector(Connector):
    @staticmethod
    def init_registry(registry: LauncherRegistry):
        registry.register_connector("ssh", SshConfiguration)

    def __init__(self, hostname: str):
        """Creates a new SSH connector

        :param hostname: a hostname specification
        """
        self.connection = Connection(hostname)
        # self.hostname = hostname
        # self.port = None

        # # FIXME: should connect on demand
        # config = paramiko.SSHConfig()
        # with Path("~/.ssh/config").expanduser().open("r") as fp:
        #     config.parse(fp)

        # lookup = config.lookup(hostname)

        # self.client = paramiko.SSHClient()

        # if "proxycommand" in lookup:
        #     paramiko.proxy.ProxyCommand(lookup["proxycommand"])

        # self.client.load_system_host_keys()
        # self.client.connect(lookup["hostname"])
        # self.sftp = self.client.open_sftp()

    @staticmethod
    def fromPath(path: SshPath):
        """Creates an SSH connector from an SshPath"""
        return SshConnector.get(path.host)

    @staticmethod
    def get(hostname):
        """Get an SSH connector from a hostname

        This method can caches SSH connectors, and is thus preferred
        to direct initialization
        """
        # TODO: cache values?
        return SshConnector(hostname)

    def __str__(self):
        return "%s" % self.hostname

    def processbuilder(self) -> ProcessBuilder:
        return SshProcessBuilder(self)

    def lock(self, path: Path, max_delay: int = -1) -> Lock:
        """Returns a lock on a file"""
        raise NotImplementedError()

    def resolve(self, path: Path, basepath: Path = None):
        raise NotImplementedError()

    def setExecutable(self, path: Path, flag: bool):
        raise NotImplementedError()

    def createtoken(self, name: str, total: int) -> Token:
        """Returns a token in the default path for the connector"""
        raise NotImplementedError()



from typing import List, TypeVar, Callable, Any
from pathlib import Path
from experimaestro import Param

from .objects import Config, LightweightTask
from .arguments import DataPath
from experimaestro import copyconfig


class SerializationLWTask(LightweightTask):
    """A serializable configuration

    This can be used to define a loading mechanism when instanciating the
    configuration
    """

    value: Param[Config]
    """The configuration that will be serialized"""

    registered: List[Config]
    """(execution only) List of configurations that use this serialized config"""

    def __post_init__(self):
        super().__post_init__()
        self.registered = []

    def register(self, config: Config):
        self.registered.append(config)


T = TypeVar("T", bound=Config)


class PathSerializationLWTask(SerializationLWTask):
    """A path based serialized configuration

    The most common case it to have
    """

    path: DataPath
    """Path containing the data"""

    @classmethod
    def construct(cls, value: T, path: Path, dep: Callable[[Config], Any]) -> T:
        value = copyconfig(value)
        return value.add_pretasks(dep(cls(value=value, path=path)))



from typing import Dict
from pathlib import Path
from experimaestro import (
    tag,
    config,
    argument,
    Config,
    Task,
    Param,
)
from experimaestro.scheduler.workspace import RunMode
from experimaestro.xpmutils import DirectoryContext


@argument("x", type=int)
@config()
class Config1:
    pass


@argument("x", type=int)
@argument("c", type=Config1)
@config()
class Config2:
    pass


def test_tag():
    c = Config1(x=5)
    c.tag("x", 5)
    assert c.tags() == {"x": 5}


def test_taggedvalue():
    c = Config1(x=tag(5))
    assert c.tags() == {"x": 5}


def test_tagcontain():
    """Test that tags are not propagated to the upper configurations"""
    c1 = Config1(x=5)
    c2 = Config2(c=c1, x=tag(3)).tag("out", 1)
    assert c1.tags() == {}
    assert c2.tags() == {"x": 3, "out": 1}


def test_inneroutput():
    class Output(Config):
        pass

    class MyTask(Task):
        mainoutput: Param[Output]
        outputs: Param[Dict[str, Output]]

    class Evaluate(Task):
        task: Param[MyTask]

    output = Output().tag("hello", "world")
    task = MyTask(outputs={}, mainoutput=output)
    task.submit(run_mode=RunMode.DRY_RUN)
    assert output.tags() == {"hello": "world"}

    output = Output().tag("hello", "world")
    task = MyTask(outputs={"a": output}, mainoutput=Output())
    task.submit(run_mode=RunMode.DRY_RUN)
    assert output.tags() == {"hello": "world"}

    evaluate = Evaluate(task=task).submit(run_mode=RunMode.DRY_RUN)
    assert evaluate.__xpm__.tags() == {"hello": "world"}


class TaskDirectoryContext(DirectoryContext):
    def __init__(self, task, path):
        super().__init__(path)
        self._task = task

    @property
    def task(self):
        return self._task


def test_objects_tags():
    """Test tags"""

    class A(Config):
        x: Param[int]

    context = DirectoryContext(Path("/__fakepath__"))
    a = A(x=tag(1))
    a.__xpm__.seal(context)
    assert a.__xpm__.tags() == {"x": 1}



# Import Python modules

import inspect
from pathlib import Path
from typing import Callable, Type as TypingType, Optional, TypeVar, Union
from sortedcontainers import SortedDict
import experimaestro.core.objects as objects
import experimaestro.core.types as types
from experimaestro.generators import PathGenerator

from .core.arguments import Argument as CoreArgument
from .core.objects import Config
from .core.types import Any, Identifier, TypeProxy, Type, ObjectType
from .utils import logger
from .checkers import Checker

# --- Annotations to define tasks and types

T = TypeVar("T")


def configmethod(method):
    """(deprecated) Annotate a method that should be kept in the configuration object"""
    return method


class config:
    """Annotations for experimaestro types"""

    def __init__(self, identifier=None, description=None):
        """[summary]

        Keyword Arguments:
            identifier {Identifier, str} -- Unique identifier of the type,
            generate by default (None)

            description {str} -- (deprecated, use
            comments) Description of the config/task, use comments with
            (default) None

            register {bool} -- False if the type should not be
            registered (debug only)

        The identifier, if not specified, will be set to `X.CLASSNAME`(by order
        of priority), where X is:
            - the parent identifier
            - the module qualified name
        """
        super().__init__()
        self.identifier = identifier
        if isinstance(self.identifier, str):
            self.identifier = Identifier(self.identifier)

        self.description = description

    def __call__(self, tp: T) -> T:
        """Annotate the class

        Depending on whether we are running or configuring,
        the behavior is different:

        - when configuring, we return a proxy class
        - when running, we return the same class

        Arguments:
            tp {type} -- The type

        Keyword Arguments:
            basetype {type} -- The base type of the class (Task or Config)
        """
        assert inspect.isclass(tp), f"{tp} is not a class"

        # Adds to xpminfo for on demand creation of information
        return ObjectType(tp, identifier=self.identifier).basetype


class Array(TypeProxy):
    """Array of object"""

    def __init__(self, type):
        self.type = Type.fromType(type)

    def __call__(self):
        return types.ArrayType(self.type)


class Choice(TypeProxy):
    """A string with a choice among several alternative"""

    def __init__(self, *args):
        self.choices = args

    def __call__(self):
        return types.StringType


class task(config):
    """Register a task"""

    def __init__(self, identifier=None, pythonpath=None, description=None):
        super().__init__(identifier, description)
        self.pythonpath = pythonpath

    def __call__(self, tp) -> type:
        # Register the type
        tp = super().__call__(tp)

        def factory(xpmtype):
            return xpmtype.getpythontaskcommand(self.pythonpath)

        tp.__getxpmtype__().taskcommandfactory = factory
        return tp


# --- argument related annotations


class param:
    """Defines an argument for an experimaestro type"""

    def __init__(
        self,
        name,
        type=None,
        default=None,
        required: bool = None,
        ignored: Optional[bool] = None,
        help: Optional[str] = None,
        checker: Optional[Checker] = None,
        constant: bool = False,
    ):
        # Determine if required
        self.name = name
        self.type = Type.fromType(type) if type else None
        self.help = help
        self.ignored = ignored
        self.default = default
        self.required = required
        self.generator = None
        self.checker = checker
        self.constant = constant

    def __call__(self, tp):
        # Don't annotate in task mode
        tp.__getxpmtype__().addAnnotation(self)
        return tp

    def process(self, xpmtype):
        # Get type from default if needed
        if self.type is None:
            if self.default is not None:
                self.type = Type.fromType(type(self.default))

        # Type = any if no type
        if self.type is None:
            self.type = types.Any

        argument = CoreArgument(
            self.name,
            self.type,
            help=self.help,
            required=self.required,
            ignored=self.ignored,
            generator=self.generator,
            default=self.default,
            checker=self.checker,
            constant=self.constant,
        )
        xpmtype.addArgument(argument)


# Just a rebind (back-compatibility)
argument = param


class option(param):
    """An argument which is ignored

    See argument
    """

    def __init__(self, *args, **kwargs):
        kwargs["ignored"] = True
        super().__init__(*args, **kwargs)


class pathoption(param):
    """Defines a an argument that will be a relative path (automatically
    set by experimaestro)"""

    def __init__(self, name: str, path=None, help=""):
        """
        :param name: The name of argument (in python)
        :param path: The relative path or a function
        """
        super().__init__(name, type=Path, help=help)

        if path is None:
            path = name

        self.generator = PathGenerator(path)


def STDERR(jobcontext, config):
    return "%s.err" % jobcontext.name


def STDOUT(jobcontext, config):
    return "%s.out" % jobcontext.name


class constant(param):
    """
    A constant argument (useful for versionning tasks)
    """

    def __init__(self, name: str, value, type=None, help=""):
        super().__init__(name, default=value, constant=True, type=type, help=help)


ConstantParam = constant


def config_only(method):
    """Marks a configuration-only method"""
    assert inspect.ismethod(method)


# --- Cache


def cache(name: str):
    """Use a cache path for a given config"""

    def annotate(method):
        return objects.cache(method, name)

    return annotate


# --- Tags


def tag(value):
    """Tag a value"""
    return objects.TaggedValue(value)


class TagDict(SortedDict):
    """A hashable dictionary"""

    def __hash__(self):
        return hash(tuple((key, value) for key, value in self.items()))

    def __setitem__(self, key, value):
        raise Exception("A tag dictionary is not mutable")


def tags(value) -> TagDict:
    """Return the tags associated with a value"""
    return TagDict(value.__xpm__.tags())


def _normalizepathcomponent(v: Any):
    if isinstance(v, str):
        return v.replace("/", "-")
    return v


def tagspath(value: Config):
    """Return a unique path made of tags and their values"""
    return "_".join(
        f"""{_normalizepathcomponent(key)}={_normalizepathcomponent(value)}"""
        for key, value in tags(value).items()
    )


# --- Deprecated


def deprecate(config: Union[TypingType[Config], Callable]):
    """Deprecate a configuration / task or
    an attribute (via a method)

    Usage:

        @deprecate
        class OldConfig(NewConfig):
            pass

        # Or only a parameter
        class MyConfig():
            @deprecate
            def oldattribute(self, value):
                # Do something with the value
                pass
    """
    if inspect.isclass(config):
        config.__getxpmtype__().deprecate()
        return config

    if inspect.isfunction(config):
        from experimaestro.core.types import DeprecatedAttribute

        return DeprecatedAttribute(config)

    raise NotImplementedError("Cannot deprecate %s", config)


def deprecateClass(klass):
    import inspect

    def __init__(self, *args, **kwargs):
        frameinfo = inspect.stack()[1]
        logger.warning(
            "Class %s is deprecated: use %s in %s:%s (%s)",
            klass.__name__,
            klass.__bases__[0].__name__,
            frameinfo.filename,
            frameinfo.lineno,
            frameinfo.code_context,
        )
        super(klass, self).__init__(*args, **kwargs)

    klass.__init__ = __init__
    return klass


def initializer(method):
    """Defines a method as an initializer that can only be called once"""

    def wrapper(self, *args, **kwargs):
        value = method(self, *args, **kwargs)
        setattr(self, method.__name__, lambda *args, **kwargs: value)

    return wrapper



# Launcher registry

from dataclasses import dataclass
import itertools
from types import new_class
from typing import ClassVar, Dict, List, Optional, Set, Type, Union
from experimaestro import Annotated
from pathlib import Path
import typing
import pkg_resources
import humanfriendly
import yaml
from yaml import Loader, Dumper
from experimaestro.utils import logger
from experimaestro.utils.yaml import (
    Initialize,
    YAMLDataClass,
    YAMLException,
    YAMLList,
    add_path_resolvers,
)

from .base import LauncherConfiguration, ConnectorConfiguration, TokenConfiguration
from .specs import CPUSpecification, CudaSpecification, HostRequirement

if typing.TYPE_CHECKING:
    from experimaestro.launchers import Launcher
    from experimaestro.tokens import Token


class LauncherNotFoundError(Exception):
    pass


@dataclass
class GPU(YAMLDataClass):
    """Represents a GPU"""

    model: str
    count: int
    memory: Annotated[int, Initialize(humanfriendly.parse_size)]

    def to_spec(self):
        return [CudaSpecification(self.memory, self.model) for _ in range(self.count)]


class GPUList(YAMLList[GPU]):
    """Represents a list of GPUs"""

    def __repr__(self):
        return f"GPUs({super().__repr__()})"

    def to_spec(self) -> List[CudaSpecification]:
        return list(itertools.chain(*[gpu.to_spec() for gpu in self]))


@dataclass
class CPU(YAMLDataClass):
    """Represents a CPU"""

    memory: Annotated[int, Initialize(humanfriendly.parse_size)] = 0
    cores: int = 1

    def to_spec(self):
        return CPUSpecification(self.memory, self.cores)


@dataclass
class Host(YAMLDataClass):
    name: str
    gpus: List[GPU]
    launchers: List[str]


Launchers = Dict[str, List[LauncherConfiguration]]
Connectors = Dict[str, Dict[str, ConnectorConfiguration]]
Tokens = Dict[str, Dict[str, TokenConfiguration]]


def new_loader(name: str) -> Type[Loader]:
    return new_class("LauncherLoader", (yaml.FullLoader,))  # type: ignore


def load_yaml(loader_cls: Type[Loader], path: Path):
    if not path.is_file():
        return None

    logger.warning(
        "Using YAML file to configure launchers is deprecated. Please remove %s using launchers.py",
        path,
    )
    logger.debug("Loading %s", path)
    with path.open("rt") as fp:
        loader = loader_cls(fp)
        try:
            return loader.get_single_data()
        finally:
            loader.dispose()


def unknown_error(loader: Loader, node):
    raise YAMLException(
        "",
        node.start_mark.name,
        node.start_mark.line,
        node.start_mark.column,
        f"No handler defined for key {node}",
    )


class LauncherRegistry:
    INSTANCES: ClassVar[Dict[Path, "LauncherRegistry"]] = {}
    CURRENT_CONFIG_DIR: ClassVar[Optional[Path]] = None

    @staticmethod
    def instance():
        """Returns an instance for the current configuration directory"""
        if LauncherRegistry.CURRENT_CONFIG_DIR is None:
            LauncherRegistry.CURRENT_CONFIG_DIR = Path(
                "~/.config/experimaestro"
            ).expanduser()

        if LauncherRegistry.CURRENT_CONFIG_DIR not in LauncherRegistry.INSTANCES:
            LauncherRegistry.INSTANCES[
                LauncherRegistry.CURRENT_CONFIG_DIR
            ] = LauncherRegistry(LauncherRegistry.CURRENT_CONFIG_DIR)

        return LauncherRegistry.INSTANCES[LauncherRegistry.CURRENT_CONFIG_DIR]

    @staticmethod
    def set_config_dir(config_dir: Path):
        LauncherRegistry.CURRENT_CONFIG_DIR = config_dir

    def __init__(self, basepath: Path):
        self.LauncherLoader: Type[Loader] = new_loader("LauncherLoader")
        self.ConnectorLoader: Type[Loader] = new_loader("ConnectorLoader")
        self.TokenLoader: Type[Loader] = new_loader("TokenLoader")
        self.Dumper: Type[Dumper] = new_class("CustomDumper", (Dumper,), {})
        self.find_launcher_fn = None

        # Add safeguards
        add_path_resolvers(
            self.LauncherLoader,
            [],
            Dict[str, LauncherConfiguration],
            dumper=self.Dumper,
        )

        # Use entry points for connectors and launchers
        for entry_point in pkg_resources.iter_entry_points("experimaestro.connectors"):
            entry_point.load().init_registry(self)

        for entry_point in pkg_resources.iter_entry_points("experimaestro.launchers"):
            entry_point.load().init_registry(self)

        for entry_point in pkg_resources.iter_entry_points("experimaestro.tokens"):
            entry_point.load().init_registry(self)

        # Register the find launcher function if it exists
        launchers_py = basepath / "launchers.py"
        if launchers_py.is_file():
            logger.info("Loading %s", launchers_py)

            from importlib import util

            spec = util.spec_from_file_location("xpm_launchers_conf", launchers_py)
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)

            self.find_launcher_fn = getattr(module, "find_launcher", None)
            if self.find_launcher_fn is None:
                logger.warn("No find_launcher() function was found in %s", launchers_py)

        # Read the configuration file
        launchers: Launchers = (
            load_yaml(self.LauncherLoader, basepath / "launchers.yaml") or {}
        )
        self.launchers = sorted(
            itertools.chain(*launchers.values()), key=lambda launcher: -launcher.weight
        )

        self.connectors: Connectors = (
            load_yaml(self.ConnectorLoader, basepath / "connectors.yaml") or {}
        )
        self.tokens: Tokens = (
            load_yaml(self.TokenLoader, basepath / "tokens.yaml") or {}
        )

    def register_launcher(self, identifier: str, cls: Type[YAMLDataClass]):
        add_path_resolvers(
            self.LauncherLoader, [identifier, None], cls, dumper=self.Dumper
        )

    def register_connector(self, identifier: str, cls: Type[YAMLDataClass]):
        add_path_resolvers(
            self.ConnectorLoader, [identifier, None], cls, dumper=self.Dumper
        )

    def register_token(self, identifier: str, cls: Type[YAMLDataClass]):
        add_path_resolvers(self.TokenLoader, [identifier], cls, dumper=self.Dumper)

    def getToken(self, identifier: str) -> "Token":
        for tokens in self.tokens.values():
            if identifier in tokens:
                return tokens[identifier].create(self, identifier)
        raise AssertionError(f"No token with identifier {identifier}")

    def getConnector(self, identifier: str):
        for connectors in self.connectors.values():
            if identifier in connectors:
                return connectors[identifier].create(self)

        # Default local connector
        if identifier == "local":
            from experimaestro.connectors.local import LocalConnector

            return LocalConnector.instance()

        raise AssertionError(f"No connector with identifier {identifier}")

    def find(
        self, *input_specs: Union[HostRequirement, str], tags: Set[str] = set()
    ) -> Optional["Launcher"]:
        """ "
        Arguments:
            spec: The processing requirements
            tags: Restrict the launchers to those containing one of the specified tags
        """

        if len(self.launchers) == 0 and self.find_launcher_fn is None:
            logger.info("No launchers.yaml file: using local host ")
            from experimaestro.launchers.direct import DirectLauncher
            from experimaestro.connectors.local import LocalConnector

            return DirectLauncher(LocalConnector.instance())

        # Parse specs
        from .parser import parse

        specs = []
        for spec in input_specs:
            if isinstance(spec, str):
                specs.extend(parse(spec))
            else:
                specs.append(spec)

        # Use launcher function
        if self.find_launcher_fn is not None:
            for spec in specs:
                if launcher := self.find_launcher_fn(spec, tags):
                    return launcher

        # We have registered launchers
        for spec in specs:
            for handler in self.launchers:
                if (not tags) or any((tag in tags) for tag in handler.tags):
                    if launcher := handler.get(self, spec):
                        return launcher
        return None


def find_launcher(
    *specs: Union[HostRequirement, str], tags: Set[str] = set()
) -> "Launcher":
    """Find a launcher matching a given specification"""
    launcher = LauncherRegistry.instance().find(*specs, tags=tags)
    if not launcher:
        raise LauncherNotFoundError(
            f"No launcher with specification: {specs}."
            "Please refer to the documentation at the following URL: "
            "https://experimaestro-python.readthedocs.io/en/latest/launchers/"
        )
    return launcher



import sys
from pathlib import Path
import time

_, notifypath, semaphorepath, *args = sys.argv
notifypath = Path(notifypath)
semaphorepath = Path(semaphorepath)

# Notify

notifypath.touch(exist_ok=False)
print("Notified", file=sys.stderr)

# ... and wait
print(f"Waiting for {semaphorepath}", file=sys.stderr)
while not semaphorepath.is_file():
    time.sleep(0.1)

print("Wait is over", file=sys.stderr)



import json
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union, TYPE_CHECKING
from experimaestro.core.context import (
    SerializationContext,
    SerializedPath,
    SerializedPathLoader,
)
from experimaestro.core.objects import ConfigInformation

if TYPE_CHECKING:
    from experimaestro import LightweightTask


def json_object(context: SerializationContext, value: Any, objects=[]):
    from experimaestro import Config

    if isinstance(value, Config):
        value.__xpm__.__get_objects__(objects, context)
    elif isinstance(value, list):
        for el in value:
            ConfigInformation.__collect_objects__(el, objects, context)
    elif isinstance(value, dict):
        for el in value.values():
            ConfigInformation.__collect_objects__(el, objects, context)
    else:
        raise NotImplementedError("Cannot serialize objects of type %s", type(value))

    # Finally, returns the output configuration
    return ConfigInformation._outputjsonvalue(value, context)


def state_dict(context: SerializationContext, obj: Any):
    """Returns a state dictionary of the object

    :param context: The serialization context
    :param obj: the object to serialize
    """
    objects = []
    data = json_object(context, obj, objects)
    return {"objects": objects, "data": data}


def save(obj: Any, save_directory: Path):
    """Saves an object into a disk file

    :param save_directory: The directory in which the object and its data will
        be saved (by default, the object is saved in "definition.json")
    """
    context = SerializationContext(save_directory=save_directory)

    data = state_dict(context, obj)
    with (save_directory / "definition.json").open("wt") as out:
        json.dump(data, out)


def get_data_loader(path: Union[str, Path, SerializedPathLoader]):
    if path is None:

        def data_loader():
            raise RuntimeError("No serialization path was given")

        return data_loader

    if callable(path):
        return path

    path = Path(path)

    def data_loader(s: Union[Path, str, SerializedPath]):
        if isinstance(s, SerializedPath):
            return path / Path(s.path)
        return path / Path(s)

    return data_loader


def from_state_dict(
    state: Dict[str, Any],
    path: Union[None, str, Path, SerializedPathLoader] = None,
    *,
    as_instance: bool = False
):
    """Load an object from a state dictionary

    :param state: The state
    :param path: A directory or a function that transforms relative file path
        into absolute ones
    :param as_instance: returns instances instead of configuration objects
    """
    objects = ConfigInformation.load_objects(
        state["objects"],
        as_instance=as_instance,
        data_loader=get_data_loader(path),
    )

    return ConfigInformation._objectFromParameters(state["data"], objects)


def load(
    path: Union[str, Path, SerializedPathLoader],
    as_instance: bool = False,
) -> Tuple[Any, List["LightweightTask"]]:
    """Load data from disk

    :param path: A directory or a function that transforms relative file path
        into absolute ones
    :param as_instance: returns instances instead of configuration objects
    """
    data_loader = get_data_loader(path)

    with data_loader("definition.json").open("rt") as fh:
        content = json.load(fh)
    return from_state_dict(content, as_instance=as_instance)


def from_task_dir(
    path: Union[str, Path, SerializedPathLoader],
    as_instance: bool = False,
):
    """Loads a task object"""
    data_loader = get_data_loader(path)
    with data_loader("params.json").open("rt") as fh:
        content = json.load(fh)

    content["data"] = {"type": "python", "value": content["objects"][-1]["id"]}

    return from_state_dict(content, as_instance=as_instance)



# Utility class that allow to avoid installing modules
# when building documentation

import sys
import re
import importlib.abc
import importlib.machinery


class Spec:
    def __init__(self, name):
        self.name = name


class Module:
    def __init__(self, loader, spec):
        self.__loader__ = loader
        self.__spec__ = spec
        self.__path__ = None
        self.__package__ = spec.name
        _, self.__name__ = f".{spec.name}".rsplit(".", 1)

    def __getattr__(self, key):
        return Module(self.__loader__, Spec(f"{self.__package__}.{key}"))

    def __repr__(self):
        return f"Fake module ({self.__package__})"


class DependencyInjectorLoader(importlib.abc.Loader):
    def create_module(self, spec):
        return Module(self, spec)

    def exec_module(self, module):
        pass


class DependencyInjectorFinder(importlib.abc.MetaPathFinder):
    def __init__(self, loader, matcher):
        self._loader = loader
        self.matcher = matcher

    def find_spec(self, fullname: str, path, target=None):
        # If we match, then do it
        if self.matcher.match(fullname):
            return self._gen_spec(fullname)

    def _gen_spec(self, fullname):
        spec = importlib.machinery.ModuleSpec(fullname, self._loader)
        return spec

    @staticmethod
    def install(matcher):
        _loader = DependencyInjectorLoader()
        _finder = DependencyInjectorFinder(_loader, matcher)
        sys.meta_path.append(_finder)



import time
from threading import Condition
from tqdm.autonotebook import tqdm
from experimaestro.scheduler import JobState, Listener
from experimaestro import Config


def jobmonitor(*outputs: Config):
    """Follow the progress of a list of jobs (in order)"""

    cv = Condition()

    class LocalListener(Listener):
        def job_state(self, job):
            with cv:
                # Just notify when something happens
                cv.notify()

    listener = LocalListener()

    for output in outputs:
        try:
            job = None
            job = output.__xpm__.job
            lastprogress = 0
            while job.scheduler is None:
                time.sleep(0.1)

            job.scheduler.addlistener(listener)

            while job.state.notstarted():
                with cv:
                    cv.wait(timeout=5000)

            # Job already completed
            if job.state.value == JobState.DONE.value:
                print("Job already completed")  # noqa: T201
            else:
                with tqdm(total=100, unit_scale=True, unit="%") as reporter:
                    progress = 0
                    if not job.state.finished():
                        while not job.state.finished():
                            progress = 0
                            if len(job.progress) > 0 and job.progress[0] is not None:
                                progress = int(job.progress[0].progress * 100)
                                delta = progress - lastprogress
                            else:
                                delta = 0

                            if delta >= 0:
                                reporter.update(delta)
                                lastprogress = progress

                            # Wait for an event
                            with cv:
                                cv.wait(timeout=5000)

                    if job.state.value == JobState.DONE.value:
                        if progress < 100:
                            reporter.update(100 - progress)
                    else:
                        raise RuntimeError(
                            f"Job did not complete successfully ({job.state.name})."
                            f"Check the error log {job.stderr}"
                        )

        finally:
            if job is not None:
                job.scheduler.removelistener(listener)



from experimaestro import argument, Config
from experimaestro.click import forwardoption
import click


def test_main():
    @argument("epochs", type=int, default=100, help="Number of learning epochs")
    class MyModel(Config):
        pass

    @forwardoption.epochs(MyModel)
    @click.command()
    def cli(epochs):
        return epochs

    epochs = cli(["--epochs", "100"], standalone_mode=False)
    assert epochs == 100


def test_rename():
    @argument("epochs", type=int, default=100, help="Number of learning epochs")
    class MyModel(Config):
        pass

    @forwardoption.epochs(MyModel, "my-epochs")
    @click.command()
    def cli(my_epochs):
        return my_epochs

    epochs = cli(["--my-epochs", "100"], standalone_mode=False)
    assert epochs == 100



# flake8: noqa: T201
import sys
from types import ModuleType
from typing import Set, Optional
import pkg_resources
from itertools import chain
from shutil import rmtree
import click
import logging
from functools import cached_property, update_wrapper
from pathlib import Path
import subprocess
from termcolor import colored, cprint

import experimaestro
from experimaestro.experiments.cli import experiments_cli
import experimaestro.launcherfinder.registry as launcher_registry

# --- Command line main options
logging.basicConfig(level=logging.INFO)


def pass_cfg(f):
    """Pass configuration information"""

    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(f, ctx.obj, *args, **kwargs)

    return update_wrapper(new_func, f)


def check_xp_path(ctx, self, path: Path):
    if not (path / ".__experimaestro__").is_file():
        cprint(f"{path} is not an experimaestro working directory", "red")
        for path in path.parents:
            if (path / ".__experimaestro__").is_file():
                cprint(f"{path} could be the folder you want", "green")
                if click.confirm("Do you want to use this folder?"):
                    return path
        sys.exit(1)

    return path


class RunConfig:
    def __init__(self):
        self.traceback = False


def pass_cfg(f):
    """Pass configuration information"""

    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        return ctx.invoke(f, ctx.obj, *args, **kwargs)

    return update_wrapper(new_func, f)


@click.group()
@click.option("--quiet", is_flag=True, help="Be quiet")
@click.option("--debug", is_flag=True, help="Be even more verbose (implies traceback)")
@click.option(
    "--traceback", is_flag=True, help="Display traceback if an exception occurs"
)
@click.pass_context
def cli(ctx, quiet, debug, traceback):
    if quiet:
        logging.getLogger().setLevel(logging.WARN)
    elif debug:
        logging.getLogger().setLevel(logging.DEBUG)

    ctx.obj = RunConfig()
    ctx.obj.traceback = traceback


# Adds the run-experiment command
cli.add_command(experiments_cli, "run-experiment")


@cli.command(help="Get version")
def version():
    print(experimaestro.__version__)


@click.argument("parameters", type=Path)
@cli.command(context_settings={"allow_extra_args": True})
def run(parameters):
    """Run a task"""

    from experimaestro.run import run as do_run

    do_run(parameters)


@click.argument("path2", type=Path)
@click.argument("path1", type=Path)
@cli.command(context_settings={"allow_extra_args": True})
def parameters_difference(path1, path2):
    """Compute the difference between two configurations"""

    from experimaestro.tools.diff import diff

    diff(path1, path2)


@click.option(
    "--clean", is_flag=True, help="Remove the socket file and its enclosing directory"
)
@click.argument("unix-path", type=Path)
@cli.command()
def rpyc_server(unix_path, clean):
    """Start an rPyC server"""
    from experimaestro.rpyc import start_server

    start_server(unix_path, clean=clean)


@cli.group()
def deprecated():
    """Manage identifier changes"""
    pass


@click.option("--fix", is_flag=True, help="Generate links to new IDs")
@click.option("--cleanup", is_flag=True, help="Remove symbolic links and move folders")
@click.argument("path", type=Path, callback=check_xp_path)
@deprecated.command(name="list")
def deprecated_list(path: Path, fix: bool, cleanup: bool):
    """List deprecated jobs"""
    from experimaestro.tools.jobs import fix_deprecated

    if cleanup and not fix:
        logging.warning("Ignoring --cleanup since we are not fixing old IDs")
    fix_deprecated(path, fix, cleanup)


@click.argument("path", type=Path, callback=check_xp_path)
@deprecated.command()
def diff(path: Path):
    """Show the reason of the identifier change for a job"""
    from experimaestro.tools.jobs import load_job
    from experimaestro import Config
    from experimaestro.core.objects import ConfigWalkContext

    _, job = load_job(path / "params.json", discard_id=False)
    _, new_job = load_job(path / "params.json")

    def check(path: str, value, new_value, done: Set[int]):
        if isinstance(value, Config):
            if id(value) in done:
                return
            done.add(id(value))

            old_id = value.__xpm__.identifier.all.hex()
            new_id = new_value.__xpm__.identifier.all.hex()

            if new_id != old_id:
                print(f"{path} differ: {new_id} vs {old_id}")

                for arg in value.__xpmtype__.arguments.values():
                    arg_value = getattr(value, arg.name)
                    arg_newvalue = getattr(new_value, arg.name)
                    check(f"{path}/{arg.name}", arg_value, arg_newvalue, done)

        elif isinstance(value, list):
            for ix, (array_value, array_newvalue) in enumerate(zip(value, new_value)):
                check(f"{path}.{ix}", array_value, array_newvalue, done)

        elif isinstance(value, dict):
            for key, dict_value in value.items():
                check(f"{path}.{key}", dict_value, new_value[key], done)

    check(".", job, new_job, set())


@click.argument("path", type=Path, callback=check_xp_path)
@click.option("--experiment", default=None, help="Restrict to this experiment")
@click.option("--tags", is_flag=True, help="Show tags")
@click.option("--ready", is_flag=True, help="Include tasks which are not yet scheduled")
@click.option("--filter", default="", help="Filter expression")
@click.option(
    "--force",
    is_flag=True,
    help="Force operation even if experiment has not been completed yet",
)
@click.option("--kill", is_flag=True, help="Kill filtered tasks (when/before running)")
@click.option("--clean", is_flag=True, help="Remove finished tasks directories")
@cli.command()
def jobs(
    path: Path,
    experiment: str,
    filter: str,
    tags: bool,
    ready: bool,
    kill: bool,
    clean: bool,
    force: bool,
):
    """Job control: list, kill and clean

    The job filter is a boolean expression where tags (alphanumeric)
    and special job information (@state for job state, @name for job full
    name) can be compared to a given value (using '~' for regex matching,
    '=', 'not in', or 'in')

    For instance,

    model = "bm25" and mode in ["a", b"] and @state = "RUNNING"

    selects jobs where the tag model is "bm25", the tag mode is either
    "a" or "b", and the state is running.

    """
    for p in (path / "xp").glob("*"):
        if experiment and p.name != experiment:
            continue

        from .filter import createFilter, JobInformation
        from experimaestro.scheduler import JobState

        _filter = createFilter(filter) if filter else lambda x: True

        print(f"* Experiment {p.name}")
        if (p / "jobs.bak").is_dir():
            cprint("  Experiment has not finished yet", "red")
            if not force and (kill or clean):
                cprint("  Preventing kill/clean (use --force if you want to)", "yellow")
                kill = False
                clean = False
        print()

        for job in p.glob("jobs/*/*"):
            info = None
            p = job.resolve()
            if p.is_dir():
                *_, scriptname = p.parent.name.rsplit(".", 1)

                info = JobInformation(p, scriptname)
                if filter:
                    if not _filter(info):
                        continue

                if info.state is None:
                    print(
                        colored(f"NODIR   {job.parent.name}/{job.name}", "red"), end=""
                    )
                elif info.state.running():
                    if kill:
                        process = info.getprocess()
                        print("KILLING", process)
                        process.kill()
                    print(
                        colored(
                            f"{info.state.name:8}{job.parent.name}/{job.name}", "yellow"
                        ),
                        end="",
                    )
                elif info.state == JobState.DONE:
                    print(
                        colored(f"DONE    {job.parent.name}/{job.name}", "green"),
                        end="",
                    )
                elif info.state == JobState.ERROR:
                    print(
                        colored(f"FAIL    {job.parent.name}/{job.name}", "red"), end=""
                    )
                else:
                    print(
                        colored(
                            f"{info.state.name:8}{job.parent.name}/{job.name}", "red"
                        ),
                        end="",
                    )

            else:
                if not ready:
                    continue
                print(colored(f"READY {job.parent.name}/{job.name}", "yellow"), end="")

            if tags:
                print(f""" {" ".join(f"{k}={v}" for k, v in info.tags.items())}""")
            else:
                print()

            if clean and info.state and info.state.finished():
                cprint("Cleaning...", "red")
                rmtree(p)
        print()


@click.option("--show-all", is_flag=True, help="Show even not orphans")
@click.option(
    "--ignore-old", is_flag=True, help="Ignore old jobs for unfinished experiments"
)
@click.option("--clean", is_flag=True, help="Prune the orphan folders")
@click.option("--size", is_flag=True, help="Show size of each folder")
@click.argument("path", type=Path, callback=check_xp_path)
@cli.command()
def orphans(path: Path, clean: bool, size: bool, show_all: bool, ignore_old: bool):
    """Check for tasks that are not part of an experimental plan"""

    jobspath = path / "jobs"

    def getjobs(path: Path):
        return ((str(p.relative_to(path)), p) for p in path.glob("*/*") if p.is_dir())

    def show(key: str, prefix=""):
        if size:
            print(
                prefix,
                subprocess.check_output(["du", "-hs", key], cwd=jobspath)
                .decode("utf-8")
                .strip(),
                sep=None,
            )
        else:
            print(prefix, key, sep=None)

    for p in (path / "xp").glob("*/jobs.bak"):
        logging.warning("Experiment %s has not completed successfully", p.parent.name)

    # Retrieve the jobs within expedriments (jobs and jobs.bak folder within experiments)
    xpjobs = set()
    if ignore_old:
        paths = (path / "xp").glob("*/jobs")
    else:
        paths = chain((path / "xp").glob("*/jobs"), (path / "xp").glob("*/jobs.bak"))

    for p in paths:
        if p.is_dir():
            for relpath, path in getjobs(p):
                xpjobs.add(relpath)

    # Now, look at stored jobs
    found = 0
    for key, jobpath in getjobs(jobspath):
        if key not in xpjobs:
            show(key)
            if clean:
                logging.info("Removing data in %s", jobpath)
                rmtree(jobpath)
        else:
            if show_all:
                show(key, prefix="[not orphan] ")
            found += 1

    print(f"{found} jobs are not orphans")


def arg_split(ctx, param, value):
    # split columns by ',' and remove whitespace
    return set(c.strip() for c in value.split(","))


@click.option("--skip", default=set(), callback=arg_split)
@click.argument("package", type=str)
@click.argument("objects", type=Path)
@cli.command()
def check_documentation(objects, package, skip):
    """Check that all the configuration and tasks are documented within a
    package, relying on the sphinx objects.inv file"""
    from experimaestro.tools.documentation import documented_from_objects, undocumented

    documented = documented_from_objects(objects)
    errors, configs = undocumented([package], documented, skip)
    for config in configs:
        cprint(f"{config.__module__}.{config.__qualname__}", "red")

    if errors > 0 or configs:
        sys.exit(1)


@click.option("--config", type=Path, help="Show size of each folder")
@click.argument("spec", type=str)
@cli.command()
def find_launchers(config: Optional[Path], spec: str):
    """Find launchers matching a specification"""
    if config is not None:
        launcher_registry.LauncherRegistry.set_config_dir(config)

    print(launcher_registry.find_launcher(spec))


class Launchers(click.MultiCommand):
    """Connectors commands"""

    @cached_property
    def commands(self):
        map = {}
        for ep in pkg_resources.iter_entry_points(f"experimaestro.{self.name}"):
            if get_cli := getattr(ep.load(), "get_cli", None):
                map[ep.name] = get_cli()
        return map

    def list_commands(self, ctx):
        return self.commands.keys()

    def get_command(self, ctx, name):
        return self.commands[name]


cli.add_command(Launchers("launchers", help="Launcher specific commands"))
cli.add_command(Launchers("connectors", help="Connector specific commands"))
cli.add_command(Launchers("tokens", help="Token specific commands"))


@cli.group()
@click.argument("workdir", type=Path, callback=check_xp_path)
@click.pass_context
def experiments(ctx, workdir):
    """Manage experiments"""
    ctx.obj = workdir


@experiments.command()
@pass_cfg
def list(workdir: Path):
    for p in (workdir / "xp").iterdir():
        if (p / "jobs.bak").exists():
            cprint(f"[unfinished] {p.name}", "yellow")
        else:
            cprint(p.name, "cyan")


def main():
    cli(obj=None)


if __name__ == "__main__":
    main()



from experimaestro.utils.asyncio import asyncThreadcheck
from .utils import logger


class Lock:
    """A lock"""

    def __init__(self):
        self._level = 0
        self.detached = False

    def detach(self):
        self.detached = True

    def acquire(self):
        if self._level == 0:
            self._level += 1
            self._acquire()
        return self

    def release(self):
        if not self.detached and self._level == 1:
            self._level -= 1
            self._release()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, *args):
        self.release()

    async def __aenter__(self):
        return await asyncThreadcheck("lock (aenter)", self.__enter__)

    async def __aexit__(self, *args):
        return await asyncThreadcheck("lock (aexit)", self.__exit__, *args)

    def _acquire(self):
        raise NotImplementedError()

    def _release(self):
        raise NotImplementedError()


class LockError(Exception):
    pass


class Locks(Lock):
    """A set of locks"""

    def __init__(self):
        super().__init__()
        self.locks = []

    def append(self, lock):
        self.locks.append(lock)

    def _acquire(self):
        for lock in self.locks:
            lock.acquire()

    def _release(self):
        logger.debug("Releasing %d locks", len(self.locks))
        for lock in self.locks:
            logger.debug("[locks] Releasing %s", lock)
            lock.release()



import time
import sys
from typing import Callable
from experimaestro import Task, pathoption
import psutil
import logging
import subprocess
import json
import signal
from experimaestro.scheduler.workspace import RunMode
from experimaestro.tests.utils import TemporaryExperiment, is_posix
from experimaestro.scheduler import JobState
from . import restart_main


def terminate(p):
    p.terminate()


def sigint(p):
    p.send_signal(signal.SIGINT)


TERMINATES_FUNC = [terminate]
if is_posix():
    TERMINATES_FUNC.append(sigint)


@pathoption("touch", "touch")
@pathoption("wait", "wait")
class Restart(Task):
    def execute(self):
        # Write the file "touch" to notify that we started
        with open(self.touch, "w") as out:
            out.write("hello")

        # Wait for the file "wait" before exiting
        while not self.wait.is_file():
            time.sleep(0.1)


def restart(terminate: Callable, experiment):
    """Check if a new experimaestro process is able to take back
    a running job

    1. Runs an experiment and kills it using "terminate" while keeping the job active
    2. Runs the same experiment
        2.1 Submit the same job
        2.2 Asserts that the job is running
        2.3 Signals to the job to end
        2.4 Asserts that the job is done

    Args:
        terminate (Callable): How to terminate the process (SIGINT / terminate)
        experiment ([type]): [description]
    """
    p = None
    xpmprocess = None
    try:
        with TemporaryExperiment("restart", maxwait=20) as xp:
            # Create the task with dry_run and so we can get the file paths
            task = Restart()
            task.submit(run_mode=RunMode.DRY_RUN)

        # Start the experiment with another process, and kill the job
        command = [
            sys.executable,
            restart_main.__file__,
            xp.workspace.path,
            experiment.__module__,
            experiment.__name__,
        ]

        logging.debug("Starting other process with: %s", command)
        xpmprocess = subprocess.Popen(command)

        counter = 0
        while not task.touch.is_file():
            time.sleep(0.1)
            counter += 1
            if counter >= 20:
                terminate(xpmprocess)
                assert False, "Timeout waiting for task to be executed"

        jobinfo = json.loads(task.__xpm__.job.pidpath.read_text())
        pid = int(jobinfo["pid"])
        p = psutil.Process(pid)

        # Now, kills experimaestro
        logging.debug("Process has started [file %s, pid %d]", task.touch, pid)
        terminate(xpmprocess)
        errorcode = xpmprocess.wait(5)
        logging.debug("Process finishing with status %d", errorcode)

        # Check that task is still running
        logging.info("Checking that job (PID %s) is still running", pid)
        assert p.is_running()

        with TemporaryExperiment("restart", maxwait=20) as xp:
            # Now, submit the job - it should pick up the process
            # where it was left
            logging.debug("Submitting the job (continues the submit)")
            job = task.__xpm__.job
            scheduler = xp.current().scheduler

            assert scheduler.submit(job) is None

            while scheduler.getJobState(job).result() == JobState.READY:
                time.sleep(0.1)

            currentState = scheduler.getJobState(job).result()
            assert (
                currentState == JobState.RUNNING
            ), f"Job is not running (state is {currentState})"

            # Notify the task
            with task.wait.open("w") as fp:
                fp.write("done")

            assert job.finalState().result() == JobState.DONE
    finally:
        # Force kill
        if xpmprocess and xpmprocess.poll() is None:
            logging.warning("Forcing to quit process %s", xpmprocess.pid)
            xpmprocess.kill()

        if p and p.is_running():
            logging.warning("Forcing to quit process %s", p.pid)
            p.terminate()



from pathlib import Path
import time
from experimaestro import Task, Param
import logging

logging.basicConfig(level=logging.INFO)


class TokenTask(Task):
    """Wait until the file is given"""

    path: Param[Path]
    """The path to watch"""

    x: Param[int]
    """A dummy parameter to create several distinct token tasks"""

    def execute(self):
        print(time.time())
        while not self.path.is_file():
            time.sleep(0.1)
        print(time.time())



import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Dict, Iterator, Optional, TypeVar, overload
import os.path
from urllib.request import urlopen
from urllib.error import HTTPError, URLError
import threading
import sys
import socket
from tqdm.auto import tqdm as std_tqdm

from .utils import logger
from experimaestro.taskglobals import Env as TaskEnv

# --- Progress and other notifications

T = TypeVar("T")


@dataclass
class LevelInformation:
    level: int
    desc: Optional[str]
    progress: float

    previous_progress: float = -1
    previous_desc: Optional[str] = None

    def modified(self, reporter: "Reporter"):
        return (
            abs(self.progress - self.previous_progress) > reporter.progress_threshold
        ) or (self.previous_desc != self.desc)

    def report(self):
        self.previous_progress = self.progress
        result = {"level": self.level, "progress": self.progress}
        if self.previous_desc != self.desc:
            self.previous_desc = self.desc
            result["desc"] = self.desc
        return result

    def __repr__(self) -> str:
        return f"[{self.level}] {self.desc} {int(self.progress*1000)/10}%"


class Reporter(threading.Thread):
    NOTIFICATION_FOLDER = ".notifications"

    console: bool
    """Whether to output to the console if no notification server is up"""

    def __init__(self, path: Path):
        """Starts a notification thread

        Arguments:
            path: The path where notification URLs will be put (one file per URL)
        """
        super().__init__(daemon=True)
        self.path = path / Reporter.NOTIFICATION_FOLDER
        self.path.mkdir(exist_ok=True)
        self.urls: Dict[str, str] = {}

        # Last check of notification URLs
        self.lastcheck = 0

        self.levels = [LevelInformation(0, None, -1)]

        self.stopping = False

        self.console = False

        self.progress_threshold = 0.01
        self.cv = threading.Condition()
        self.start()

    def stop(self):
        self.stopping = True
        with self.cv:
            self.cv.notifyAll()

    @staticmethod
    def isfatal_httperror(e: Exception) -> bool:
        """Returns True if this HTTP error indicates that the server won't recover"""
        if isinstance(e, HTTPError):
            if e.code >= 400 and e.code < 500:
                return True
        elif isinstance(e, URLError):
            if isinstance(e.reason, ConnectionRefusedError):
                return True
            if isinstance(e.reason, socket.gaierror) and e.reason.errno == -2:
                return True

        return False

    def modified(self):
        return any(level.modified(self) for level in self.levels)

    def check_urls(self):
        mtime = os.path.getmtime(self.path)
        if mtime > self.lastcheck:
            for f in self.path.iterdir():
                self.urls[f.name] = f.read_text().strip()
                logger.info("Added new notification URL: %s", self.urls[f.name])
                f.unlink()

            self.lastcheck = os.path.getmtime(self.path)

    def run(self):
        logger.info("Running notification thread")

        while True:
            with self.cv:
                self.cv.wait_for(lambda: self.stopping or self.modified())
                if self.stopping:
                    break

            # Notify (out of the CV locking)
            toremove = []

            # Check if new notification servers are on
            self.check_urls()

            if self.urls:
                # OK, let's go
                for level in self.levels:
                    if level.modified(self):
                        params = level.report()

                        # Go over all URLs
                        for key, baseurl in self.urls.items():
                            url = "{}/progress?{}".format(
                                baseurl, urllib.parse.urlencode(params)
                            )
                            logger.debug("Reporting progress %s", params)
                            try:
                                with urlopen(url) as _:
                                    logger.debug(
                                        "Notification send for %s [%s]",
                                        baseurl,
                                        level,
                                    )
                            except Exception as e:
                                logger.warning(
                                    "Progress: %s [error while notifying %s]: %s",
                                    level,
                                    url,
                                    e,
                                )
                                if Reporter.isfatal_httperror(e):
                                    toremove.append(key)

                # Removes unvalid URLs
                for key in toremove:
                    logger.info("Removing notification URL %s", self.urls[key])
                    del self.urls[key]
            elif self.console:
                for level in self.levels:
                    if level.modified(self):
                        params = level.report()
                        logger.info("Progress: %s", level)

    def eoj(self):
        with self.cv:
            self.check_urls()
            if self.urls:
                # Go over all URLs
                for key, baseurl in self.urls.items():
                    url = "{}?status=eoj".format(baseurl)
                    try:
                        with urlopen(url) as _:
                            logger.debug(
                                "EOJ botification sent for %s",
                                baseurl,
                            )
                    except Exception:
                        logger.warning(
                            "Could not report EOJ",
                        )

    def set_progress(
        self, progress: float, level: int, desc: Optional[str], console=False
    ):
        """Sets the new progress if sufficiently different"""
        with self.cv:
            if (
                (level + 1) != len(self.levels)
                or (progress != self.levels[level].progress)
                or (desc != self.levels[level].desc)
            ):
                self.console = console
                self.levels = self.levels[: (level + 1)]
                while level >= len(self.levels):
                    self.levels.append(LevelInformation(level, None, 0.0))
                if desc:
                    self.levels[level].desc = desc
                self.levels[level].progress = progress

                self.cv.notify_all()

    INSTANCE: ClassVar[Optional["Reporter"]] = None

    @staticmethod
    def instance():
        if Reporter.INSTANCE is None:
            taskpath = TaskEnv.instance().taskpath
            assert taskpath is not None, "Task path is not defined"
            Reporter.INSTANCE = Reporter(taskpath)
        return Reporter.INSTANCE


def progress(value: float, level=0, desc: Optional[str] = None, console=False):
    """When called from a running task, report the progress

    Args:
        level: The level (starting from 0)
        value: The current value
        desc: An optional description of the current task
    """
    if TaskEnv.instance().slave:
        # Skip if in a slave process
        return
    Reporter.instance().set_progress(value, level, desc, console=console)


def report_eoj():
    """Notify that the job is over"""
    Reporter.instance().eoj()


class xpm_tqdm(std_tqdm):
    """XPM wrapper for experimaestro that automatically reports progress to the
    server"""

    def __init__(self, iterable=None, file=None, *args, **kwargs):
        # Report progress bar
        # newprogress(title=, pos=abs(self.pos))
        _file = file or sys.stderr
        self.is_tty = hasattr(_file, "isatty") or _file.isatty()

        super().__init__(iterable, disable=False, file=file, *args, **kwargs)
        progress(0.0, level=self.pos, desc=kwargs.get("desc", None), console=False)

    def update(self, n=1):
        result = super().update(n)
        if self.total is not None:
            progress(self.n / self.total, level=self.pos, console=False)
        return result

    def refresh(self, nolock=False, lock_args=None):
        if self.is_tty:
            super().refresh(nolock=nolock, lock_args=lock_args)


@overload
def tqdm(**kwargs) -> xpm_tqdm:
    ...


@overload
def tqdm(iterable: Optional[Iterator[T]] = None, **kwargs) -> Iterator[T]:
    ...


def tqdm(*args, **kwargs):
    return xpm_tqdm(*args, **kwargs)  # type: ignore



from contextlib import contextmanager
from pathlib import Path
from typing import Union
from importlib import resources
from experimaestro.compat import cached_property


class ResourcePathWrapper:
    """Simple wrapper for resource path"""

    def __init__(self, path: Path):
        self.path = path

    @staticmethod
    def create(package: str, path: Path):
        return ResourcePathWrapper(Path(*package.split(".")) / path)

    def __truediv__(self, subpath: Union[str, Path]):
        return ResourcePathWrapper(self.path / subpath)

    @cached_property
    def package(self):
        parents = [s.name for s in reversed(self.path.parents)][1:]
        return ".".join(parents)

    @property
    def name(self):
        return self.path.name

    def is_file(self):
        return resources.is_resource(self.package, self.name)

    @contextmanager
    def open(self, *args, **kwargs):
        with resources.path(self.package, self.name) as path:
            yield path.open(*args, **kwargs)



from typing import Any, Callable, Dict, List, Optional
import pyparsing as pp
from pathlib import Path
import json
from experimaestro.compat import cached_property
import regex
from experimaestro.scheduler import JobState


class JobInformation:
    def __init__(self, path: Path, scriptname: str):
        self.path = path
        self.scriptname = scriptname

    @cached_property
    def params(self):
        return json.loads((self.path / "params.json").read_text())

    @cached_property
    def tags(self) -> List[str]:
        return self.params["tags"]

    @cached_property
    def state(self) -> Optional[JobState]:
        if (self.path / f"{self.scriptname}.pid").is_file():
            return JobState.RUNNING
        elif (self.path / f"{self.scriptname}.done").is_file():
            return JobState.DONE
        elif (self.path / f"{self.scriptname}.failed").is_file():
            return JobState.ERROR
        else:
            return None

    def getprocess(self):
        from experimaestro.connectors import Process
        from experimaestro.connectors.local import LocalConnector

        connector = LocalConnector.instance()
        pinfo = json.loads((self.path / f"{self.scriptname}.pid").read_text())
        return Process.fromDefinition(connector, pinfo)


# --- classes for processing


class VarExpr:
    def __init__(self, values):
        (self.varname,) = values

    def get(self, info: JobInformation):
        if self.varname == "@state":
            return info.state.name if info.state else None

        if self.varname == "@name":
            return str(info.path.parent.name)

        return info.tags.get(self.varname, None)

    def __repr__(self):
        return f"""VAR<{self.varname}>"""


class BaseInExpr:
    def __init__(self, values):
        self.var, *stringList = values
        self.values = set(stringList)


class InExpr(BaseInExpr):
    def filter(self, information: JobInformation):
        value = self.var.get(information)
        return value in self.values

    def __repr__(self):
        return f"""IN<{self.varname}|{",".join(self.values)}>"""


class NotInExpr(BaseInExpr):
    def filter(self, information: JobInformation):
        value = self.var.get(information)
        return value not in self.values

    def __repr__(self):
        return f"""NOT_IN<{self.varname}|{",".join(self.values)}>"""


class RegexExpr:
    def __init__(self, tokens):
        self.var, expr = tokens
        self.regex = regex.compile(expr)

    def __repr__(self):
        return f"""REGEX[{self.varname}, {self.value}]"""

    def matches(self, manager, publication):
        if self.varname == "tag":
            return self.value in publication.tags

        raise AssertionError()

    def filter(self, information: JobInformation):
        value = self.var.get(information)
        if not value:
            return False

        return self.regex.match(value)


class ConstantString:
    def __init__(self, tokens):
        (self.value,) = tokens

    def get(self, information: JobInformation):
        return self.value

    def __repr__(self):
        return f'"{self.value}"'


class EqExpr:
    def __init__(self, tokens):
        self.var1, self.var2 = tokens

    def __repr__(self):
        return f"""EQ[{self.var1}, {self.var2}]"""

    def filter(self, information: JobInformation):
        return self.var1.get(information) == self.var2.get(information)


class LogicExpr:
    """Logical expression"""

    def __init__(self, tokens):
        self.operator, self.y = tokens
        self.x = None

    def filter(self, information: JobInformation):
        if self.operator == "and":
            return self.y.filter(information) and self.x.filter(information)

        return self.y.filter(information) or self.x.filter(information)

    @staticmethod
    def summary(tokens):
        if len(tokens) == 1:
            return tokens[0]
        v = tokens[1]
        v.x = tokens[0]
        for token in tokens[2:]:
            token.x = v
            v = token
        return v

    @staticmethod
    def generator(tokens):
        if len(tokens) == 1:
            return tokens[0]
        v = LogicExpr(("and", tokens[1]))
        v.x = tokens[0]
        return v

    def __repr__(self):
        return f"""{self.x} [{self.operator}] {self.y}"""


# --- Grammar

l = pp.Literal

lpar, rpar, lbra, rbra, eq, comma, pipe, tilde = map(pp.Suppress, "()[]=,|~")
quotedString = pp.QuotedString('"', unquoteResults=True) | pp.QuotedString(
    "'", unquoteResults=True
)

var = l("@state") | l("@name") | pp.Word(pp.alphas)
var.setParseAction(VarExpr)

regexExpr = var + tilde + quotedString
regexExpr.setParseAction(RegexExpr)

varQuotedString = quotedString
varQuotedString.setParseAction(ConstantString)
eqExpr = var + eq + (var | quotedString)
eqExpr.setParseAction(EqExpr)

stringList = quotedString + pp.ZeroOrMore(comma + quotedString)

notInExpr = var + (pp.Suppress(l("not in")) + lbra + stringList + rbra)
notInExpr.setParseAction(NotInExpr)

inExpr = var + (pp.Suppress(l("in")) + lbra + stringList + rbra)
inExpr.setParseAction(InExpr)

matchExpr = eqExpr | regexExpr | inExpr | notInExpr

booleanOp = l("and") | l("or")
logicExpr = (
    matchExpr + pp.ZeroOrMore((booleanOp + matchExpr).setParseAction(LogicExpr))
).setParseAction(LogicExpr.summary)
parenExpr = logicExpr | (lpar + logicExpr + rpar)

filterExpr = (
    parenExpr + pp.ZeroOrMore((booleanOp + parenExpr).setParseAction(LogicExpr))
).setParseAction(LogicExpr.summary)
expr = (matchExpr + pp.Optional(pipe + filterExpr)).setParseAction(LogicExpr.generator)


def createFilter(query: str) -> Callable[[Dict[str, Any]], bool]:
    """Returns a filter object given a query"""
    (r,) = logicExpr.parseString(query, parseAll=True)
    return r.filter



from collections import ChainMap
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Optional
from experimaestro.settings import WorkspaceSettings, Settings


class RunMode(str, Enum):
    NORMAL = "normal"
    """Normal run"""

    GENERATE_ONLY = "generate"
    """Do not run, but generate the params.json file"""

    DRY_RUN = "dry-run"
    """Do not run"""


class Workspace:
    """An experimental workspace

    This workspace is created by an experiment object and is used by launchers
    to set up jobs
    """

    CURRENT = None
    settings: "Settings"
    worspace: "WorkspaceSettings"

    """Creates a workspace for experiments"""

    def __init__(
        self,
        settings: "Settings",
        workspace_settings: "WorkspaceSettings",
        launcher=None,
        run_mode: RunMode = None,
    ):
        self.settings = settings
        self.workspace_settings = workspace_settings

        path = self.workspace_settings.path
        self.notificationURL: Optional[str] = None
        if isinstance(path, Path):
            path = path.absolute()
        self.path = path
        self.run_mode = run_mode
        from ..launchers import Launcher

        self.launcher = launcher or Launcher.get(path)

        self.env = ChainMap({}, workspace_settings.env, settings.env)

    def __enter__(self):
        self.old_workspace = Workspace.CURRENT
        Workspace.CURRENT = self

    def __exit__(self, *args):
        Workspace.CURRENT = self.old_workspace

    @cached_property
    def alt_workspaces(self):
        for ws_id in self.workspace_settings.alt_workspaces:
            yield self.settings.workspaces[ws_id]

    @property
    def alt_workdirs(self):
        yield from map(lambda ws: ws.path, self.workspace_settings.alt_workspaces)

    @property
    def connector(self):
        """Returns the default connector"""
        return self.launcher.connector

    @property
    def jobspath(self):
        """Folder for jobs"""
        return self.path / "jobs"

    @property
    def experimentspath(self):
        """Folder for experiments"""
        return self.path / "xp"

    @property
    def configcachepath(self):
        """Folder for jobs"""
        return self.path / "config"



import os
from omegaconf import OmegaConf, SCMode
from dataclasses import field, dataclass
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, List


@dataclass
class ServerSettings:
    port: Optional[int] = None
    """Port for the server"""

    host: Optional[str] = None
    """Hostname for the server"""

    autohost: Optional[str] = None
    """Automatic hostname: values can be `fqdn` or `name`"""

    token: Optional[str] = None
    """Token for the server"""


@dataclass
class WorkspaceSettings:
    """Defines the workspace"""

    id: str
    """The workspace identifier"""

    path: Path = field()
    """The workspace path"""

    env: Dict[str, str] = field(default_factory=dict)
    """Workspace specific environment variables"""

    alt_workspaces: List[str] = field(default_factory=list)
    """Alternative workspaces to find jobs or experiments"""

    def __post_init__(self):
        self.path = self.path.expanduser().resolve()


@dataclass
class Settings:
    server: ServerSettings = field(default_factory=ServerSettings)
    workspaces: List[WorkspaceSettings] = field(default_factory=list)

    env: Dict[str, str] = field(default_factory=dict)
    """Default environment variables"""


@lru_cache()
def get_settings(path: Optional[Path] = None) -> Settings:
    if "PYTEST_CURRENT_TEST" in os.environ:
        return Settings()
    else:
        schema = OmegaConf.structured(Settings)

        path = path or Path("~/.config/experimaestro/settings.yaml").expanduser()
        if not path.is_file():
            return schema.to_object()

        conf = OmegaConf.load(path)
        return OmegaConf.to_container(
            OmegaConf.merge(schema, conf), structured_config_mode=SCMode.INSTANTIATE
        )


def get_workspace(id: Optional[str] = None) -> Optional[WorkspaceSettings]:
    """Return the workspace settings given an id (or None for the default one)"""
    workspaces = get_settings().workspaces
    if workspaces:
        if id is None:
            return workspaces[0]
        for workspace in workspaces:
            if id == workspace.id:
                return workspace

    return None



from mypy.plugin import Plugin


class ExperimaestroPlugin(Plugin):
    """Just do nothing for now"""

    def get_class_decorator_hook(self, tada):
        pass

    def get_customize_class_mro_hook(self, tada):
        pass


def plugin(version: str):
    return ExperimaestroPlugin



from pathlib import Path
import marshmallow as mm


class JsonSettings:
    @classmethod
    def load(cls, path: Path):
        if path.is_file():
            settings = cls.SCHEMA().loads(path.read_text())
        else:
            settings = cls()
        settings.path = path
        return settings

    def save(self):
        """Save the preferences"""
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(self.SCHEMA().dumps(self))


class PathField(mm.fields.Field):
    """Field that serializes to a title case string and deserializes
    to a lower case string.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        return Path(value)

    def _deserialize(self, value, attr, data, **kwargs):
        return str(value.absolute())



from itertools import chain
from typing import Dict, Any
from pathlib import Path
import json
from termcolor import colored


class ObjectProxy:
    id: str
    module: str
    type: str
    identifier: str

    ignored: set[str]
    ignore: bool
    fields: Dict[str, Any]

    def __init__(self, data: Dict[str, Any]):
        self.id = data["id"]
        self.module = data["module"]
        self.type = data["type"]
        self.identifier = data["identifier"]
        self.ignore = data.get("ignore", False)
        self.ignored = set(data.get("ignored", []))
        self.fields = {}

    def __repr__(self):
        return f"{self.module}.{self.type}({self.identifier})"


Store = Dict[str, ObjectProxy]


def build_value(data, store: Store):
    if isinstance(data, dict):
        t = data.get("type", None)
        if t == "python":
            return store[data["value"]]
        elif t == "path":
            # Do not return anything
            return None  # Path(data["value"])
        elif t == "path.serialized":
            # Again, do not return anything (won't change the identifier)
            return None
        elif t is None:
            return {key: build_value(value, store) for key, value in data.items()}
        assert False, f"Data type {t} not handled"
    elif isinstance(data, list):
        return [build_value(value, store) for value in data]

    return data


def build_object(data: Dict[str, Any], store: Store) -> ObjectProxy:
    object = ObjectProxy(data)
    store[data["id"]] = object

    for key, value in data["fields"].items():
        object.fields[key] = build_value(value, store)

    return object


def load(path: Path):
    data = json.loads(path.read_text())
    store = {}
    for definition in data["objects"]:
        object = build_object(definition, store)

    return object


def print_diff(path: str, conf1: Any, conf2: Any):
    if type(conf1) != type(conf2):
        print(f"[{colored(path, 'red')}] {conf1} and {conf2} of different types")

    if isinstance(conf1, ObjectProxy) and isinstance(conf2, ObjectProxy):
        if conf1.identifier != conf2.identifier:
            print(
                f"[{colored(path, 'yellow')}] {conf1} and {conf2} have different identifiers"
            )
            keys = set(chain(conf1.fields.keys(), conf2.fields.keys()))
            for key in keys:
                print_diff(
                    f"{path}.{key}" if path else key,
                    conf1.fields.get(key, None),
                    conf2.fields.get(key, None),
                )

    elif isinstance(conf1, dict) and isinstance(conf2, dict):
        keys = set(chain(conf1.keys(), conf2.keys()))
        for key in keys:
            print_diff(
                f"{path}.{key}" if path else key,
                conf1.get(key, None),
                conf2.get(key, None),
            )

    elif isinstance(conf1, list) and isinstance(conf2, list):
        if len(conf1) != len(conf2):
            print(
                f"[{colored(path, 'red')}] {conf1} and {conf2} have not the same length"
            )
        else:
            for i in range(len(conf1)):
                print_diff(f"{path}.{i}" if path else key, conf1[i], conf2[i])

    elif conf1 != conf2:
        print(f"[{colored(path, 'red')}] {conf1} and {conf2} are not equal")


def diff(path1: Path, path2: Path):
    conf1, conf2 = load(path1), load(path2)

    print_diff("", conf1, conf2)



from pathlib import Path
import pytest
from experimaestro.launchers import Launcher
from experimaestro.tests.utils import TemporaryExperiment
from .common import waitFromSpec, takeback


@pytest.mark.timeout(3)
def test_local_process(tmp_path: Path):
    launcher = Launcher.get(tmp_path)
    waitFromSpec(tmp_path, launcher)


@pytest.mark.timeout(3)
def test_local_takeback(tmp_path):
    """Test whether a task can be taken back when running"""
    txp1 = TemporaryExperiment("slurm-takeback-1", workdir=tmp_path / "xp")
    txp2 = TemporaryExperiment("slurm-takeback-2", workdir=tmp_path / "xp")
    datapath = tmp_path / "data"
    launcher = Launcher.get(tmp_path)

    takeback(launcher, datapath, txp1, txp2)



# flake8: noqa: F401

from pathlib import Path

# Annotations
from .annotations import (
    config,
    task,
    param,
    ConstantParam,
    constant,
    option,
    pathoption,
    cache,
    Identifier,
    Array,
    TagDict,
    tag,
    tags,
    tagspath,
    STDOUT,
    STDERR,
    deprecate,
    # deprecated
    argument,
    initializer,
    # Method
    config_only,
)
from .core.serialization import load, save, state_dict, from_state_dict, from_task_dir
from .core.arguments import (
    # Types
    Param,
    Option,
    Meta,
    DataPath,
    Annotated,
    Constant,
    # Annotations helpers
    help,
    default,
)
from .generators import pathgenerator
from .core.objects import (
    Config,
    copyconfig,
    setmeta,
    Task,
    LightweightTask,
    ObjectStore,
)
from .core.context import SerializationContext
from .core.serializers import SerializationLWTask, PathSerializationLWTask
from .core.types import Any, SubmitHook
from .launchers import Launcher
from .scheduler.workspace import Workspace, RunMode
from .scheduler import Scheduler, experiment, FailedExperiment
from .notifications import progress, tqdm
from .checkers import Choices
from .xpmutils import DirectoryContext
from .mkdocs.annotations import documentation
from .scheduler.base import Job
from .launcherfinder.registry import LauncherRegistry


def set_launcher(launcher: Launcher):
    Workspace.CURRENT.launcher = launcher


# Get version
__version__ = "0.0.0"
__version_tuple__ = (0, 0, 0)



"""Command line parsing"""

import os
from pathlib import Path
import signal
import sys
import json
from typing import List
import fasteners
from experimaestro.notifications import progress, report_eoj
from .core.types import ObjectType
from experimaestro.utils import logger
from experimaestro.core.objects import ConfigInformation
import experimaestro.taskglobals as taskglobals
import atexit


def parse_commandline(argv=None):
    """Called when executing a task"""
    if argv is None:
        argv = sys.argv[1:]
    taskid, params = argv
    tasktype = ObjectType.REGISTERED[taskid]
    with open(params, "r") as fp:
        params = json.load(fp)
        ConfigInformation.LOADING = True
        task = tasktype(**params)
        ConfigInformation.LOADING = False
        task.execute()


def run(parameters: Path):
    with open(parameters, "r") as fp:
        params = json.load(fp)
        env = taskglobals.Env.instance()

        env.wspath = Path(params["workspace"])
        env.taskpath = parameters.parent

        task = ConfigInformation.fromParameters(params["objects"])
        task.__taskdir__ = Path.cwd()

        # Set the tags
        task.__tags__ = params["tags"]

        # Notify that the task has started
        progress(0)

        # Execute the task
        task.execute()


def rmfile(path: Path):
    if path.is_file():
        logger.debug("Removing file %s", path)
        path.unlink()


class TaskRunner:
    """Runs a task, after locking"""

    def __init__(self, scriptpath: str, lockfiles: List[str]):
        # Sets the working directory
        self.scriptpath = Path(scriptpath)
        self.pidfile = self.scriptpath.with_suffix(".pid")
        self.lockfiles = lockfiles
        self.donepath = self.scriptpath.with_suffix(".done")
        self.failedpath = self.scriptpath.with_suffix(".failed")
        self.started = False
        self.locks = []
        env = taskglobals.Env.instance()
        env.taskpath = self.scriptpath.parent

        self.cleaned = False

    def cleanup(self):
        if not self.cleaned:
            self.cleaned = True
            logger.info("Cleaning up")
            rmfile(self.pidfile)
            for lock in self.locks:
                try:
                    if lock.acquired:
                        logger.info("Releasing lock")
                        lock.release()
                        logger.info("Released lock")
                except Exception:
                    logger.error("Error while releasing lock %s", lock)

            if self.started:
                report_eoj()
            logger.info("Finished cleanup")

    def handle_error(self, code, frame_type):
        logger.info("Error handler: finished with code %d", code)
        self.failedpath.write_text(str(code))
        self.cleanup()
        logger.info("Exiting")
        sys.exit(1)

    def run(self):
        atexit.register(self.cleanup)
        sigterm_handler = signal.signal(signal.SIGTERM, self.handle_error)
        sigint_handler = signal.signal(signal.SIGINT, self.handle_error)

        def remove_signal_handlers(remove_cleanup=True):
            """Removes cleanup in forked processes"""
            signal.signal(signal.SIGTERM, sigterm_handler)
            signal.signal(signal.SIGINT, sigint_handler)
            atexit.unregister(self.cleanup)

        if sys.platform != "win32":
            os.register_at_fork(after_in_child=remove_signal_handlers)

        try:
            workdir = self.scriptpath.parent
            os.chdir(workdir)
            os.getpid()
            logger.info("Working in directory %s", workdir)

            for lockfile in self.lockfiles:
                fullpath = str(Path(lockfile).resolve())
                logger.info("Locking %s", fullpath)
                lock = fasteners.InterProcessLock(fullpath)
                # MAYBE: should have a clever way to lock
                # Problem = slurm would have a job doing nothing...
                # Fix = maybe with two files
                if not lock.acquire(blocking=True):
                    logger.info("Could not lock %s", lockfile)
                    raise AssertionError("Could not lock %s", lockfile)
                self.locks.append(lock)

            # Check if failed/done have been generated by another job
            if self.donepath.is_file():
                logger.info("Job already completed")
            else:
                logger.info("Running task")
                rmfile(self.failedpath)
                self.started = True
                run(workdir / "params.json")

                # ... remove the handlers
                logger.info("Task ended successfully")
                remove_signal_handlers(remove_cleanup=False)

                # Everything went OK
                sys.exit(0)
        except Exception:
            logger.exception("Got exception while running")
            self.handle_error(1, None)

        except SystemExit as e:
            if e.code == 0:
                # Normal exit, just create the ".done" file
                self.donepath.touch()

                # ... and finish the exit process
                logger.info("This is the end (TODO: remove this line)")
                raise
            else:
                self.handle_error(e.code, None)



import argparse
import sys

parser = argparse.ArgumentParser()
parser.add_argument("--fail", action="store_true")
parser.add_argument("--nodes", type=int)
parser.add_argument("--time", type=int)
parser.add_argument("--gpus-per-node", type=int)
parser.add_argument("--gpus", type=int)
opts = parser.parse_args()

if opts.fail:
    sys.exit(1)

for key, value in vars(opts).items():
    if key != "fail" and value is not None:
        print(key, value, sep="=", file=sys.stdout)



from pathlib import Path
import sys
import time
import logging
import asyncio
from experimaestro.scheduler import JobState
from experimaestro.connectors import Process, Redirect
from experimaestro.launchers import Launcher
from experimaestro import Task, Param

logger = logging.getLogger("xpm.tests.launchers")


def waitFromSpec(tmp_path: Path, launcher: Launcher):
    builder = launcher.processbuilder()

    started = tmp_path / "started"
    semaphore = tmp_path / "semaphore"
    scriptfile = Path(__file__).parent / ".." / "scripts" / "notifyandwait.py"
    assert scriptfile.exists()

    builder.command = [
        sys.executable,
        scriptfile,
        started,
        semaphore,
    ]
    builder.detach = True
    builder.stdout = Redirect.file(tmp_path / "stdout")
    builder.stderr = Redirect.file(tmp_path / "stderr")

    logging.info("Starting job")
    process = builder.start()  # type: BatchSlurmProcess
    spec = process.tospec()
    logging.info("Got job ID %s", spec["pid"])

    while not started.exists():
        time.sleep(0.1)

    logging.info("Job started")

    restored = Process.fromDefinition(launcher.connector, spec)

    assert asyncio.run(restored.aio_isrunning()), "Process is not running"

    logging.info("Waiting for job to end")
    semaphore.touch()
    restored.wait()

    # Should not be running anymore
    assert not asyncio.run(restored.aio_isrunning()), "Process is running"

    return restored


# --- Take back code


class WaitUntilTouched(Task):
    touching: Param[Path]
    waiting: Param[Path]

    def execute(self):
        self.touching.touch()

        while not self.waiting.exists():
            time.sleep(0.01)


def takeback(launcher, datapath, txp1, txp2):
    datapath.mkdir()
    touching = datapath / "touching"
    waiting = datapath / "waiting"

    with txp1 as xp1:  # flake8: noqa: F841
        WaitUntilTouched(touching=touching, waiting=waiting).submit(launcher=launcher)

        logger.debug("Waiting for task to create 'touching' file")
        while not touching.is_file():
            time.sleep(0.01)

        with txp2 as xp2:  # flake8: noqa: F841
            result = WaitUntilTouched(touching=touching, waiting=waiting).submit(
                launcher=launcher
            )

            logger.debug("Waiting for job to be in running mode")
            while result.__xpm__.job.state != JobState.RUNNING:
                time.sleep(0.1)

            waiting.touch()



from experimaestro.connectors.ssh import SshPath

# --- Test SSH path and SSH path manipulation


def test_absolute():
    path = SshPath("ssh://host//a/path")
    assert path.host == "host"
    assert path.is_absolute()


def test_relative():
    path = SshPath("ssh://host")
    assert path.host == "host"
    assert path.hostpath == ""
    assert path._parts == []
    assert not path.is_absolute()


def test_relative_withpath():
    path = SshPath("ssh://host/relative/path")
    assert path.host == "host"
    assert path.hostpath == "relative/path"
    assert not path.is_absolute()


def test_relative_absolute():
    path = SshPath("ssh://host") / "/absolute/path"
    assert path.host == "host"
    assert path.hostpath == "/absolute/path"
    assert path.is_absolute()


def test_relative_compose():
    path = SshPath("ssh://host/abc") / "relative/path"
    assert path.host == "host"
    assert path.hostpath == "abc/relative/path"
    assert not path.is_absolute()
